#!/usr/bin/env python
import re
from typing import Any
from typing import Callable
from typing import Dict

from paasta_tools.utils import paasta_print


ConstraintState = Dict[str, Dict[str, Any]]
ConstraintOp = Callable[[str, str, str, ConstraintState], bool]


def max_per(constraint_value, offer_value, attribute, state: ConstraintState):
    if not constraint_value:
        constraint_value = 1
    state_value = state.get('MAX_PER', {}).get(attribute, {}).get(offer_value, 0)
    return state_value <= int(constraint_value)


# lambda arg: [constraint value, offer value, attribute, state]
# example constraint: ['pool', 'MAX_PER', 5]
#   constraint value: 5
#   offer value:      default
#   attribute:        pool
#   state:            {'MAX_PER' => {'pool' => {'default' => 6}}}
CONS_OPS: Dict[str, ConstraintOp] = {
    'EQUALS': lambda cv, ov, *_: cv == ov,
    'LIKE': lambda cv, ov, *_: bool(re.match(cv, ov)),
    'UNLIKE': lambda cv, ov, *_: not(re.match(cv, ov)),
    'MAX_PER': max_per,
    'UNIQUE': max_per,
}


def nested_inc(op, _, attr_val, attr_name, state, step=1):
    """Increments relevant counter by step from args array"""
    oph = state.setdefault(op, {})
    nameh = oph.setdefault(attr_name, {})
    nameh.setdefault(attr_val, 0)
    nameh[attr_val] += step
    return state


# lambda args same as CONS_OPS + update step
UPDATE_OPS = {
    'EQUALS': lambda *_: None,
    'LIKE': lambda *_: None,
    'UNLIKE': lambda *_: None,
    'MAX_PER': lambda *args: nested_inc('MAX_PER', *args),
    'UNIQUE': lambda *args: nested_inc('MAX_PER', *args),
}


def check_offer_constraints(offer, constraints, state):
    """Returns True if all constraints are satisfied by offer's attributes,
    returns False otherwise. Prints a error message and re-raises if an error
    was thrown."""
    for (attr, op, val) in constraints:
        try:
            offer_attr = next(
                (x for x in offer.attributes if x.name == attr), None,
            )
            if offer_attr is None:
                paasta_print("Attribute not found for a constraint: %s" % attr)
                return False
            elif not(CONS_OPS[op](val, offer_attr.text.value, offer_attr.name, state)):
                paasta_print("Constraint not satisfied: [{} {} {}] for {} with {}".format(
                    attr, op, val, offer_attr.text.value, state,
                ))
                return False
        except Exception as err:
            paasta_print("Error while matching constraint: [{} {} {}] {}".format(
                attr, op, val, str(err),
            ))
            raise err

    return True


def update_constraint_state(offer, constraints, state, step=1):
    """Mutates state for each offer attribute found in constraints by calling
    relevant UPDATE_OP lambda"""
    for (attr, op, val) in constraints:
        for oa in offer.attributes:
            if attr == oa.name:
                UPDATE_OPS[op](val, oa.text.value, attr, state, step)



# Copyright 2015-2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import argparse
import csv
import datetime
import logging
import re
from collections import defaultdict
from time import sleep
from urllib.parse import urlsplit

import chronos
import dateutil
import isodate
import pytz
import service_configuration_lib
from croniter import croniter
from crontab import CronSlices

from paasta_tools import monitoring_tools
from paasta_tools.mesos_tools import get_mesos_network_for_net
from paasta_tools.mesos_tools import mesos_services_running_here
from paasta_tools.secret_tools import get_secret_hashes
from paasta_tools.tron_tools import parse_time_variables
from paasta_tools.utils import deep_merge_dictionaries
from paasta_tools.utils import DEFAULT_SOA_DIR
from paasta_tools.utils import get_config_hash
from paasta_tools.utils import get_service_instance_list
from paasta_tools.utils import get_services_for_cluster
from paasta_tools.utils import InstanceConfig
from paasta_tools.utils import InvalidJobNameError
from paasta_tools.utils import load_system_paasta_config
from paasta_tools.utils import load_v2_deployments_json
from paasta_tools.utils import NoConfigurationForServiceError
from paasta_tools.utils import NoDeploymentsAvailable
from paasta_tools.utils import paasta_print
from paasta_tools.utils import PaastaColors
from paasta_tools.utils import PaastaNotConfiguredError
from paasta_tools.utils import time_cache
from paasta_tools.utils import timeout


# In Marathon spaces are not allowed, in Chronos periods are not allowed.
# In the Chronos docs a space is suggested as the natural separator
SPACER = " "
# Until Chronos supports dots in the job name, we use this separator internally
INTERNAL_SPACER = '.'
# Chronos creates Mesos tasks with an id composed of some arbitrary strings,
# the app's full name, a spacer, and a timestamp. This variable is that
# spacer. Note that we don't control this spacer, i.e. you can't change it
# here and expect the world to change with you. We need to know what it is so
# we can decompose Mesos task ids.
MESOS_TASK_SPACER = ':'

# this is the identifier prepended to a job name for use
# with temporary jobs, like those created by ``chronos_rerun.py``
TMP_JOB_IDENTIFIER = "tmp"

VALID_BOUNCE_METHODS = ['graceful']
EXECUTION_DATE_FORMAT = "%Y-%m-%dT%H:%M:%S"

log = logging.getLogger(__name__)

# we never want to hear from this - this means
# that callers don't have to worry about setting this
logging.getLogger("crontab").setLevel(logging.CRITICAL)


# Variable used by DFS algorithm
_ENTERED, _ENTERED_AND_EXITED = range(0, 2)


class LastRunState:
    """Cheap enum to represent the state of the last run"""
    Success, Fail, NotRun = range(0, 3)


class JobType:
    """ An enum to represent job types """
    Dependent, Scheduled = range(0, 2)


class ChronosNotConfigured(Exception):
    pass


class InvalidParentError(Exception):
    pass


class ChronosConfig(dict):

    def __init__(self, config):
        super().__init__(config)

    def get_url(self):
        """:returns: The Chronos API endpoint"""
        try:
            return self['url']
        except KeyError:
            raise ChronosNotConfigured('Could not find chronos url in system chronos config')

    def get_username(self):
        """:returns: The Chronos API username"""
        try:
            return self['user']
        except KeyError:
            raise ChronosNotConfigured('Could not find chronos user in system chronos config')

    def get_password(self):
        """:returns: The Chronos API password"""
        try:
            return self['password']
        except KeyError:
            raise ChronosNotConfigured('Could not find chronos password in system chronos config')


def load_chronos_config():
    try:
        return ChronosConfig(load_system_paasta_config().get_chronos_config())
    except PaastaNotConfiguredError:
        raise ChronosNotConfigured("Could not find chronos_config in configuration directory")


class CachingChronosClient(chronos.ChronosClient):

    @time_cache(ttl=20)
    def list(self):
        return super().list()

    @time_cache(ttl=20)
    def metrics(self):
        return super().metrics()

    @time_cache(ttl=20)
    def scheduler_graph(self):
        return super().scheduler_graph()


def get_chronos_client(config, cached=False):
    """Returns a chronos client object for interacting with the API"""
    chronos_hosts = config.get_url()
    chronos_hostnames = [urlsplit(hostname).netloc for hostname in chronos_hosts]
    log.info("Attempting to connect to Chronos servers: %s" % chronos_hosts)
    if cached:
        return CachingChronosClient(
            servers=chronos_hostnames,
            username=config.get_username(),
            password=config.get_password(),
            scheduler_api_version=None,
        )
    else:
        return chronos.connect(
            servers=chronos_hostnames,
            username=config.get_username(),
            password=config.get_password(),
            scheduler_api_version=None,
        )


def compose_job_id(service, instance):
    """In chronos, a job is just service{SPACER}instance """
    return f"{service}{SPACER}{instance}"


def decompose_job_id(job_id):
    """ A custom implementation of utils.decompose_job_id, accounting for the
    possibility of TMP_JOB_IDENTIFIER being prepended to the job name """
    decomposed = job_id.split(SPACER)
    if len(decomposed) == 3:
        if not decomposed[0].startswith(TMP_JOB_IDENTIFIER):
            raise InvalidJobNameError('invalid job id %s' % job_id)
        else:
            return (decomposed[1], decomposed[2])
    elif len(decomposed) == 2:
        return (decomposed[0], decomposed[1])
    else:
        raise InvalidJobNameError('invalid job id %s' % job_id)


class InvalidChronosConfigError(Exception):
    pass


@time_cache(ttl=30)
def read_chronos_jobs_for_service(service, cluster, soa_dir=DEFAULT_SOA_DIR):
    chronos_conf_file = 'chronos-%s' % cluster
    return service_configuration_lib.read_extra_service_information(
        service,
        chronos_conf_file,
        soa_dir=soa_dir,
    )


def load_chronos_job_config(
    service: str,
    instance: str,
    cluster: str,
    load_deployments: bool = True,
    soa_dir: str = DEFAULT_SOA_DIR,
) -> 'ChronosJobConfig':
    general_config = service_configuration_lib.read_service_configuration(
        service,
        soa_dir=soa_dir,
    )

    if instance.startswith('_'):
        raise InvalidJobNameError(
            f"Unable to load chronos job config for {service}.{instance} as instance name starts with '_'",
        )
    service_chronos_jobs = read_chronos_jobs_for_service(service, cluster, soa_dir=soa_dir)
    if instance not in service_chronos_jobs:
        raise NoConfigurationForServiceError(f'No job named "{instance}" in config file chronos-{cluster}.yaml')
    branch_dict = None
    general_config = deep_merge_dictionaries(overrides=service_chronos_jobs[instance], defaults=general_config)

    if load_deployments:
        deployments_json = load_v2_deployments_json(service, soa_dir=soa_dir)
        temp_instance_config = ChronosJobConfig(
            service=service,
            cluster=cluster,
            instance=instance,
            config_dict=general_config,
            branch_dict=None,
            soa_dir=soa_dir,
        )
        branch = temp_instance_config.get_branch()
        deploy_group = temp_instance_config.get_deploy_group()
        branch_dict = deployments_json.get_branch_dict(service, branch, deploy_group)

    return ChronosJobConfig(
        service=service,
        cluster=cluster,
        instance=instance,
        config_dict=general_config,
        branch_dict=branch_dict,
        soa_dir=soa_dir,
    )


class ChronosJobConfig(InstanceConfig):
    config_filename_prefix = 'chronos'

    def __init__(self, service, instance, cluster, config_dict, branch_dict, soa_dir=DEFAULT_SOA_DIR):
        super().__init__(
            cluster=cluster,
            instance=instance,
            service=service,
            config_dict=config_dict,
            branch_dict=branch_dict,
            soa_dir=soa_dir,
        )

    def get_service(self):
        return self.service

    def get_job_name(self):
        return self.instance

    def get_owner(self):
        overrides = self.get_monitoring()
        return monitoring_tools.get_team(overrides=overrides, service=self.get_service())

    def get_bounce_method(self):
        """Returns the bounce method specified for the Chronos job.

        Options are:
        * ``graceful``: disables the old version but allows it to finish its current run
        * ``brutal``: disables the old version and immediately kills any running tasks it has
        If unspecified, defaults to ``graceful``.
        """
        return self.config_dict.get('bounce_method', 'graceful')

    def get_env(self):
        """The expected input env for PaaSTA is a dictionary of key/value pairs
        Chronos requires an array of dictionaries in a very specific format:
        https://mesos.github.io/chronos/docs/api.html#sample-job"""
        original_env = super().get_env()
        return [{"name": key, "value": value} for key, value in original_env.items()]

    def get_calculated_constraints(self, system_paasta_config):
        constraints = self.get_constraints()
        if constraints is not None:
            return constraints
        else:
            constraints = self.get_extra_constraints()
            constraints.extend(
                self.get_deploy_constraints(
                    blacklist=self.get_deploy_blacklist(),
                    whitelist=self.get_deploy_whitelist(),
                    system_deploy_blacklist=system_paasta_config.get_deploy_blacklist(),
                    system_deploy_whitelist=system_paasta_config.get_deploy_whitelist(),
                ),
            )
            constraints.extend(self.get_pool_constraints())
        assert isinstance(constraints, list)
        return [[str(val) for val in constraint] for constraint in constraints]

    def check_bounce_method(self):
        bounce_method = self.get_bounce_method()
        if bounce_method not in VALID_BOUNCE_METHODS:
            return False, ('The specified bounce method "%s" is invalid. It must be one of (%s).'
                           % (bounce_method, ', '.join(VALID_BOUNCE_METHODS)))
        return True, ''

    def get_epsilon(self):
        return self.config_dict.get('epsilon', 'PT60S')

    def get_retries(self):
        return self.config_dict.get('retries', 2)

    def get_disabled(self):
        return self.config_dict.get('disabled', False)

    def get_schedule(self):
        return self.config_dict.get('schedule')

    def get_schedule_interval_in_seconds(self, seconds_ago=0):
        """Return the job interval in seconds or None if there is no interval

        :params seconds_ago: return an interval the job had in the past
        """
        schedule = self.get_schedule()
        if schedule is None:
            return None

        if CronSlices.is_valid(schedule):
            try:
                job_tz = pytz.timezone(self.get_schedule_time_zone())
            except (pytz.exceptions.UnknownTimeZoneError, AttributeError):
                job_tz = pytz.utc
            c = croniter(schedule, datetime.datetime.now(job_tz) - datetime.timedelta(seconds=seconds_ago))
            return c.get_next() - c.get_prev()
        else:
            try:
                _, _, interval = self.get_schedule().split('/')
                return int(float(isodate.parse_duration(interval).total_seconds()))
            except (isodate.ISO8601Error, ValueError):
                return None

    def get_schedule_time_zone(self):
        return self.config_dict.get('schedule_time_zone')

    def get_parents(self):
        parents = self.config_dict.get('parents', [])
        if parents is not None and not isinstance(parents, list):
            return [parents]
        else:
            return parents

    def get_shell(self):
        """Per https://mesos.github.io/chronos/docs/api.html, ``shell`` defaults
        to true, but if arguments are set, they will be ignored. If arguments are
        set in our config, then we need to set shell: False so that they will
        activate."""
        args = self.get_args()
        return args == [] or args is None

    def check_epsilon(self):
        epsilon = self.get_epsilon()
        try:
            isodate.parse_duration(epsilon)
        except isodate.ISO8601Error:
            return False, 'The specified epsilon value "%s" does not conform to the ISO8601 format.' % epsilon
        return True, ''

    def check_retries(self):
        retries = self.get_retries()
        if retries is not None:
            if not isinstance(retries, int):
                return False, 'The specified retries value "%s" is not a valid int.' % retries
        return True, ''

    def check_parents(self):
        parents = self.get_parents()
        if parents is not None:
            results = [(parent, check_parent_format(parent)) for parent in parents]
            badly_formatted = [res[0] for res in results if res[1] is False]
            if len(badly_formatted) > 0:
                return False, ('The job name(s) %s is not formatted correctly: expected service.instance'
                               % ", ".join(badly_formatted))
        return True, ''

    def check_cmd(self):
        command = self.get_cmd()
        if command:
            try:
                parse_time_variables(command=command)
                return True, ""
            except (ValueError, KeyError, TypeError):
                return False, ("Unparseable command '%s'. Hint: do you need to escape {} chars?") % command
        else:
            return True, ""

    # a valid 'repeat_string' is 'R' or 'Rn', where n is a positive integer representing the number of times to repeat
    # more info: https://en.wikipedia.org/wiki/ISO_8601#Repeating_intervals
    def _check_schedule_repeat_helper(self, repeat_string):
        pattern = re.compile(r'^R\d*$')
        return pattern.match(repeat_string) is not None

    def check_schedule(self):
        msgs = []
        schedule = self.get_schedule()

        if schedule is not None:
            if not CronSlices.is_valid(schedule):
                try:
                    repeat, start_time, interval = schedule.split('/')  # the parts have separate validators
                except ValueError:
                    return (
                        False, (
                            'The specified schedule "%s" is neither a valid cron schedule nor a valid'
                            ' ISO 8601 schedule' % schedule
                        ),
                    )

                # an empty start time is not valid ISO8601 but Chronos accepts it: '' == current time
                if start_time == '':
                    msgs.append('The specified schedule "%s" does not contain a start time' % schedule)
                else:
                    # Check if start time contains time zone information
                    try:
                        dt = isodate.parse_datetime(start_time)
                        if not hasattr(dt, 'tzinfo'):
                            msgs.append('The specified start time "%s" must contain a time zone' % start_time)
                    except isodate.ISO8601Error as exc:
                        msgs.append('The specified start time "%s" in schedule "%s" does '
                                    'not conform to the ISO 8601 format:\n%s' % (start_time, schedule, str(exc)))

                parsed_interval = None
                try:
                    # 'interval' and 'duration' are interchangeable terms
                    parsed_interval = isodate.parse_duration(interval)
                except isodate.ISO8601Error:
                    msgs.append('The specified interval "%s" in schedule "%s" '
                                'does not conform to the ISO 8601 format.' % (interval, schedule))

                # don't allow schedules more frequent than every minute we have
                # to be careful here, since the isodate library returns
                # different datatypes according to whether there is a
                # yearly/monthly period (and if that year or month period is
                # 0).  unfortunately, the isodate library *is* okay with you
                # specifying fractional and negative periods. Chronos's parser
                # will barf at a fractional period, but be okay with a negative
                # one, so if someone does try to do something like "R1//P0.01M"
                # then the API request to upload the job will fail.  TODO:
                # detect when someone is trying to add a fractional period?
                if (parsed_interval and isinstance(parsed_interval, datetime.timedelta) and
                        parsed_interval < datetime.timedelta(seconds=60)):
                    msgs.append('Unsupported interval "%s": jobs must be run at an interval of > 60 seconds' % interval)

                if not self._check_schedule_repeat_helper(repeat):
                    msgs.append('The specified repeat "%s" in schedule "%s" '
                                'does not conform to the ISO 8601 format.' % (repeat, schedule))

        return len(msgs) == 0, '\n'.join(msgs)

    # TODO we should use pytz for best possible tz info and validation
    # TODO if tz specified in start_time, compare to the schedule_time_zone and warn if they differ
    # TODO if tz not specified in start_time, set it to time_zone
    # NOTE confusingly, the accepted time zone format for 'schedule_time_zone' is different than in 'schedule'!
    # 'schedule_time_zone': tz database format (https://en.wikipedia.org/wiki/List_of_tz_database_time_zones)
    # 'schedule': ISO 8601 format (https://en.wikipedia.org/wiki/ISO_8601#Time_zone_designators)
    # TODO maybe we don't even want to support this a a separate parameter? instead require it to be specified
    # as a component of the 'schedule' parameter?
    def check_schedule_time_zone(self):
        time_zone = self.get_schedule_time_zone()
        if time_zone is not None:
            return True, ''
            # try:
            # TODO validate tz format
            # except isodate.ISO8601Error as exc:
            #     return False, ('The specified time zone "%s" does not conform to the tz database format:\n%s'
            #                    % (time_zone, str(exc)))
        return True, ''

    def check(self, param):
        check_methods = {
            'bounce_method': self.check_bounce_method,
            'epsilon': self.check_epsilon,
            'retries': self.check_retries,
            'cpus': self.check_cpus,
            'mem': self.check_mem,
            'disk': self.check_disk,
            'schedule': self.check_schedule,
            'scheduleTimeZone': self.check_schedule_time_zone,
            'security': self.check_security,
            'dependencies_reference': self.check_dependencies_reference,
            'parents': self.check_parents,
            'cmd': self.check_cmd,
            'deploy_group': self.check_deploy_group,
        }
        supported_params_without_checks = ['description', 'owner', 'disabled']
        if param in check_methods:
            return check_methods[param]()
        elif param in supported_params_without_checks:
            return True, ''
        else:
            return False, 'Your Chronos config specifies "%s", an unsupported parameter.' % param

    def format_chronos_job_dict(self, docker_url, docker_volumes, docker_cfg_location, constraints):
        net = get_mesos_network_for_net(self.get_net())
        command = parse_time_variables(self.get_cmd()) if self.get_cmd() else self.get_cmd()

        complete_config = {
            'name': self.get_job_name(),
            'container': {
                'image': docker_url,
                'network': net,
                'type': 'DOCKER',
                'volumes': docker_volumes,
                'parameters': self.format_docker_parameters(),
            },
            'uris': [docker_cfg_location],
            'environmentVariables': self.get_env(),
            'mem': self.get_mem(),
            'cpus': self.get_cpus(),
            'disk': self.get_disk(),
            'constraints': constraints,
            'command': command,
            'arguments': self.get_args(),
            'epsilon': self.get_epsilon(),
            'retries': self.get_retries(),
            'async': False,  # we don't support async jobs
            'disabled': self.get_disabled(),
            'owner': self.get_owner(),
            'scheduleTimeZone': self.get_schedule_time_zone(),
            'shell': self.get_shell(),
        }
        if self.get_schedule() is not None:
            complete_config['schedule'] = self.get_schedule()
        else:
            # The input to parents is the normal paasta syntax, but for chronos we have to
            # convert it to what chronos expects, which uses its own spacer
            complete_config["parents"] = [paasta_to_chronos_job_name(parent) for parent in self.get_parents()]
        return complete_config

    # 'docker job' requirements: https://mesos.github.io/chronos/docs/api.html#adding-a-docker-job
    def validate(self):
        error_msgs = []
        # Use InstanceConfig to validate shared config keys like cpus and mem
        error_msgs.extend(super().validate())

        for param in [
            'epsilon', 'retries', 'cpus', 'mem', 'disk',
            'schedule', 'scheduleTimeZone', 'parents', 'cmd',
            'security', 'dependencies_reference',
        ]:
            check_passed, check_msg = self.check(param)
            if not check_passed:
                error_msgs.append(check_msg)

        return len(error_msgs) == 0, error_msgs

    def get_healthcheck_mode(self, _):
        # Healthchecks are not supported yet in chronos
        return None

    def get_desired_state_human(self):
        desired_state = self.get_desired_state()
        if desired_state == 'start':
            return PaastaColors.bold('Scheduled')
        elif desired_state == 'stop':
            return PaastaColors.bold('Disabled')
        else:
            return PaastaColors.red('Unknown (desired_state: %s)' % desired_state)

    def get_nerve_namespace(self):
        """Currently not implemented yet, as we don't have the code
        to announce chronos tasks in nerve"""
        return None


def paasta_to_chronos_job_name(job_name):
    """Converts a paasta-style job name service.main to what chronos
    will have in its api as the job name: "service foo" """
    return job_name.replace(INTERNAL_SPACER, SPACER)


def list_job_names(service, cluster=None, soa_dir=DEFAULT_SOA_DIR):
    """A chronos-specific wrapper around utils.get_service_instance_list.

    :param name: The service name
    :param cluster: The cluster to read the configuration for
    :param soa_dir: The SOA config directory to read from
    :returns: A list of tuples of (name, job) for each job defined for the service name
    """
    return get_service_instance_list(service, cluster, 'chronos', soa_dir)


def get_chronos_jobs_for_cluster(cluster=None, soa_dir=DEFAULT_SOA_DIR):
    """A chronos-specific wrapper around utils.get_services_for_cluster

    :param cluster: The cluster to read the configuration for
    :param soa_dir: The SOA config directory to read from
    :returns: A list of tuples of (service, job_name)"""
    return get_services_for_cluster(cluster, 'chronos', soa_dir)


def create_complete_config(service, job_name, soa_dir=DEFAULT_SOA_DIR):
    """Generates a complete dictionary to be POST'ed to create a job on Chronos"""
    system_paasta_config = load_system_paasta_config()
    chronos_job_config = load_chronos_job_config(
        service, job_name, system_paasta_config.get_cluster(), soa_dir=soa_dir,
    )
    docker_url = chronos_job_config.get_docker_url()
    docker_volumes = chronos_job_config.get_volumes(system_volumes=system_paasta_config.get_volumes())

    constraints = chronos_job_config.get_calculated_constraints(
        system_paasta_config=system_paasta_config,
    )

    complete_config = chronos_job_config.format_chronos_job_dict(
        docker_url=docker_url,
        docker_volumes=docker_volumes,
        docker_cfg_location=system_paasta_config.get_dockercfg_location(),
        constraints=constraints,
    )

    complete_config['name'] = compose_job_id(service, job_name)

    # resolve conflicts between the 'desired_state' and soa_configs disabled
    # flag.
    desired_state = chronos_job_config.get_desired_state()
    soa_disabled_state = complete_config['disabled']

    resolved_disabled_state = determine_disabled_state(
        desired_state,
        soa_disabled_state,
    )
    complete_config['disabled'] = resolved_disabled_state

    config_for_hash = get_config_for_bounce_hash(
        complete_config=complete_config,
        service=service,
        soa_dir=soa_dir,
        system_paasta_config=system_paasta_config,
    )
    # we use the undocumented description field to store a hash of the chronos config.
    # this makes it trivial to compare configs and know when to bounce.
    complete_config['description'] = get_config_hash(
        config=config_for_hash,
        force_bounce=chronos_job_config.get_force_bounce(),
    )

    log.debug("Complete configuration for instance is: %s" % complete_config)
    return complete_config


def get_config_for_bounce_hash(complete_config, service, soa_dir, system_paasta_config):
    environment_variables = {x['name']: x['value'] for x in complete_config['environmentVariables']}
    secret_hashes = get_secret_hashes(
        environment_variables=environment_variables,
        secret_environment=system_paasta_config.get_vault_environment(),
        service=service,
        soa_dir=soa_dir,
    )
    if secret_hashes:
        complete_config['paasta_secrets'] = secret_hashes
    return complete_config


def determine_disabled_state(branch_state, soa_disabled_state):
    """Determine the *real* disabled state for a job. There are two sources of truth for
    the disabled/enabled state of a job: the 'disabled' flag in soa-configs,
    and the state set in deployments.json by `paasta stop` or `paasta start`.
    If the two conflict, then this determines the 'real' state for it.

    We take a defensive stance here: if either source has disabled as True,
    then we go with that - better safe than sorry I guess?

    Note that this means that `paasta start` and `paasta emergency-start` are noop if you
    have 'disabled:True' in your soa-config.

    :param branch_state: the desired state on the branch - either start or
        stop.
    :param soa_disabled_state: the value of the disabled state in soa-configs.
    """
    if branch_state not in ['start', 'stop']:
        raise ValueError('Expected branch_state to be start or stop')

    if branch_state == 'start' and soa_disabled_state is True:
        return True
    elif branch_state == 'stop' and soa_disabled_state is False:
        return True
    else:
        return soa_disabled_state


def _safe_parse_datetime(dt):
    """
    Parse a datetime, swallowing exceptions.

    :param dt: A string containing a datetime
    :returns: A datetime.datetime object representing ``dt``. If a datetime
    string is unparseable, it is represented as a datetime.datetime set to the
    epoch in UTC (i.e. a value which will never be the most recent).
    """
    epoch = datetime.datetime(1970, 1, 1, tzinfo=dateutil.tz.tzutc())
    try:
        parsed_dt = isodate.parse_datetime(dt)
    # I tried to limit this to isodate.ISO8601Error but parse_datetime() can
    # also throw "AttributeError: 'NoneType' object has no attribute 'split'",
    # and presumably other exceptions.
    except Exception as exc:
        log.debug("Failed to parse datetime '%s'" % dt)
        log.debug(exc)
        parsed_dt = epoch
    return parsed_dt


def cmp_datetimes(first, second):
    """Compare two datetime strings and sort by most recent.

    :param first: A string containing a datetime
    :param second: A string containing a datetime
    :returns: -1 if ``first`` is more recent, 1 if ``second`` is more recent, or 0
        if they are equivalent.
    """
    parsed_first = _safe_parse_datetime(first)
    parsed_second = _safe_parse_datetime(second)
    if parsed_first > parsed_second:
        return -1
    elif parsed_first == parsed_second:
        return 0
    else:
        return 1


def filter_enabled_jobs(jobs):
    """Given a list of chronos jobs, find those which are not disabled"""
    return [job for job in jobs if job['disabled'] is False]


def last_success_for_job(job):
    """
    Given a job, find the last time it was successful. In the case that it hasn't
    completed a successful run, return None.
    """
    return job.get('lastSuccess', None)


def last_failure_for_job(job):
    """
    Given a job, find the last time it failed. In the case that it has never failed,
    return None.
    """
    return job.get('lastError', None)


def get_status_last_run(job):
    """
    Return the time of the last run of a job and the appropriate LastRunState.
    """
    last_success = last_success_for_job(job)
    last_failure = last_failure_for_job(job)
    if not last_success and not last_failure:
        return (None, LastRunState.NotRun)
    elif not last_failure:
        return (last_success, LastRunState.Success)
    elif not last_success:
        return (last_failure, LastRunState.Fail)
    else:
        if cmp_datetimes(last_success, last_failure) <= 0:
            return (last_success, LastRunState.Success)
        else:
            return (last_failure, LastRunState.Fail)


def get_job_type(job):
    """ Given a job, return a chronos_tools.JobType """
    if job.get('schedule') is not None:
        return JobType.Scheduled
    elif job.get('parents') is not None:
        return JobType.Dependent
    else:
        raise ValueError('Expected either schedule field or parents field')


def sort_jobs(jobs):
    """Takes a list of chronos jobs and returns a sorted list where the job
    with the most recent result is first.

    :param jobs: list of dicts of job configuration, as returned by the chronos client
    """
    def get_key(job):
        failure = last_failure_for_job(job)
        success = last_success_for_job(job)
        newest = failure if cmp_datetimes(failure, success) < 0 else success
        return _safe_parse_datetime(newest)

    return sorted(
        jobs,
        key=get_key,
        reverse=True,
    )


def get_chronos_status_for_job(client, service, instance):
    """
    Returns the status (queued, idle, running, etc.) for a specific job from
    the undocumented Chronos csv api that has the following format:

    node,example_job,success,queued
    node,parent_job,success,idle
    node,child_job,success,idle
    link,parent_job,child_job

    :param client: Chronos client object
    :param service: service name
    :param instance: instance name
    """
    resp = client.scheduler_graph()
    lines = csv.reader(resp.splitlines())
    for line in lines:
        if line[0] == "node" and line[1] == compose_job_id(service, instance):
            return line[3]


def lookup_chronos_jobs(client, service=None, instance=None, include_disabled=False, include_temporary=False):
    """Discovers Chronos jobs and filters them with ``filter_chronos_jobs()``.

    :param client: Chronos client object
    :param service: passed on to ``filter_chronos_jobs()``
    :param instance: passed on to ``filter_chronos_jobs()``
    :param include_disabled: passed on to ``filter_chronos_jobs()``
    :returns: list of job dicts discovered by ``client`` and filtered by
        ``filter_chronos_jobs()`` using the other parameters
    """
    jobs = client.list()
    return filter_chronos_jobs(
        jobs=jobs,
        service=service,
        instance=instance,
        include_disabled=include_disabled,
        include_temporary=include_temporary,
    )


def filter_non_temporary_chronos_jobs(jobs):
    """
    Given a list of Chronos jobs, as pulled from the API, remove those
    which are defined as 'temporary' jobs.

    :param jobs: a list of chronos jobs
    :returns: a list of chronos jobs containing the same jobs as those provided
        by the ``jobs`` parameter, but with temporary jobs removed.
    """
    return [job for job in jobs if not job['name'].startswith(TMP_JOB_IDENTIFIER)]


def filter_chronos_jobs(jobs, service, instance, include_disabled, include_temporary):
    """Filters a list of Chronos jobs based on several criteria.

    :param jobs: a list of jobs, as calculated in ``lookup_chronos_jobs()``
    :param service: service we're looking for. If None, don't filter based on this key.
    :param instance: instance we're looking for. If None, don't filter based on this key.
    :param include_disabled: boolean indicating if disabled jobs should be
        included in the returned list
    :returns: list of job dicts whose name matches the arguments (if any)
        provided
    """
    matching_jobs = []
    for job in jobs:
        try:
            job_service, job_instance = decompose_job_id(job['name'])
        except InvalidJobNameError:
            continue
        if (
            (service is None or job_service == service) and
            (instance is None or job_instance == instance)
        ):
            if job['disabled'] and not include_disabled:
                continue
            else:
                matching_jobs.append(job)
    if include_temporary is True:
        return matching_jobs
    else:
        return filter_non_temporary_chronos_jobs(matching_jobs)


@timeout()
def wait_for_job(client, job_name):
    """Wait for a job to launch.

    :param client: The Chronos client
    :param job_name: The name of the job to wait for
    """
    found = False
    while not found:
        found = job_name in [job['name'] for job in client.list()]
        if found:
            return True
        else:
            paasta_print("waiting for job %s to launch. retrying" % (job_name))
            sleep(0.5)


def parse_execution_date(execution_date_string):
    """Parses execution_date_string into a datetime object.
    Can be used as an ArgumentParser `type=` hint.
    :param execution_date_string: string
    :return: a datetime.datetime instance
    :raise: ArgumentTypeError
    """
    try:
        parsed = datetime.datetime.strptime(execution_date_string, EXECUTION_DATE_FORMAT)
    except ValueError:
        raise argparse.ArgumentTypeError('must be in the format "%s"' % EXECUTION_DATE_FORMAT)
    return parsed


def uses_time_variables(chronos_job):
    """
    Given a chronos job config, returns True the if the command uses time variables
    """
    current_command = chronos_job.get_cmd()
    interpolated_command = parse_time_variables(
        command=current_command,
        parse_time=datetime.datetime.utcnow(),
    )
    return current_command != interpolated_command


def check_parent_format(parent):
    """ A predicate defining if the parent field string is formatted correctly """
    return len(parent.split(".")) == 2


def disable_job(client, job):
    job["disabled"] = True
    log.debug("Disabling job: %s" % job)
    client.update(job)


def delete_job(client, job):
    log.debug("Deleting job: %s" % job["name"])
    client.delete(job["name"])


def create_job(client, job):
    log.debug("Creating job: %s" % job)
    client.add(job)


def update_job(client, job):
    log.debug("Updating job: %s" % job)
    client.update(job)


def get_jobs_for_service_instance(service, instance, include_disabled=True, include_temporary=False):
    """A wrapper for lookup_chronos_jobs that includes building the necessary chronos client.
    :param service: the service to be queried
    :param instance: the instance to be queried
    :param include_disabled: include disabled jobs in the search. default: True
    :param include_temporary: include temporary in the search. default: False
    :returns: the jobs for a given service instance
    """
    chronos_config = load_chronos_config()
    matching_jobs = lookup_chronos_jobs(
        client=get_chronos_client(chronos_config),
        service=service,
        instance=instance,
        include_disabled=include_disabled,
        include_temporary=include_temporary,
    )
    return sort_jobs(matching_jobs)


def compose_check_name_for_service_instance(check_name, service, instance):
    """Compose a sensu check name for a given job"""
    return f'{check_name}.{service}{INTERNAL_SPACER}{instance}'


def get_temporary_jobs_for_service_instance(client, service, instance):
    """ Given a service and instance, find any temporary jobs
    for that job, as created by chronos_rerun.
    """
    temporary_jobs = []
    all_jobs = lookup_chronos_jobs(
        client=client,
        service=service,
        instance=instance,
        include_disabled=True,
        include_temporary=True,
    )
    for job in all_jobs:
        if job['name'].startswith(TMP_JOB_IDENTIFIER):
            temporary_jobs.append(job)
    return temporary_jobs


def is_temporary_job(job):
    """Assert whether a job is a temporary one, identifiable
    by it's name starting with TMP_JOB_IDENTIFIER
    :param job: the chronos job to check
    :returns: a boolean indicating if a job is a temporary job
    """
    return job['name'].startswith(TMP_JOB_IDENTIFIER)


def chronos_services_running_here():
    """See what chronos services are being run by a mesos-slave on this host.
    :returns: A list of triples of (service, instance, port)"""

    return mesos_services_running_here(
        framework_filter=lambda fw: fw['name'].startswith('chronos'),
        parse_service_instance_from_executor_id=lambda task_id: decompose_job_id(task_id.split(MESOS_TASK_SPACER)[3]),
    )


ENTERED, ENTERED_AND_EXITED = range(0, 2)


def dfs(node, neighbours_mapping, ignore_cycles=False, node_status=None):
    """
    Execute Deep-First search on a graph starting from a specific `node`.

    `neighbours_mapping` contains the graph links as adjacency list
        neighbours_mapping[x] = [y, z] means that the node `x` as two outgoing edges/links directed to `y` and `z`

    :param node: starting node for DFS execution
    :param neighbours_mapping: graph adjacency list
    :type neighbours_mapping: dict
    :param ignore_cycles: raise an exception if a cycle is identified in the graph
    :type ignore_cycles: bool
    :param node_status: dictionary that keep tracks of the already visited nodes.
        NOTE: this is meant for internal usage only, please do NOT set it during method call
    :type node_status: dict

    :return: list of visited nodes sorted by exit order
    :type: list
    """
    visited_nodes = []

    if node_status is None:
        node_status = {}

    if node_status.get(node) == _ENTERED_AND_EXITED:
        return []

    node_status[node] = _ENTERED
    for current_node in neighbours_mapping[node]:
        current_node_state = node_status.get(current_node, None)
        if current_node_state == _ENTERED:
            if ignore_cycles:
                continue
            else:
                raise ValueError("cycle")
        if current_node_state == _ENTERED_AND_EXITED:
            continue

        visited_nodes.extend(
            dfs(
                node=current_node,
                neighbours_mapping=neighbours_mapping,
                node_status=node_status,
                ignore_cycles=ignore_cycles,
            ),
        )

    visited_nodes.append(node)
    node_status[node] = _ENTERED_AND_EXITED
    return visited_nodes


@time_cache(ttl=30)
def _get_related_jobs_and_configs(cluster, soa_dir=DEFAULT_SOA_DIR):
    """
    For all the Chronos jobs defined in cluster, extract the list of topologically sorted related Chronos.
    Two individual jobs are considered related each other if exists a dependency relationship between them.

    :param cluster: cluster from which extracting the jobs
    :return: tuple(related jobs mapping, jobs configuration)
    """
    chronos_configs = {}
    for service in service_configuration_lib.read_services_configuration(soa_dir=soa_dir):
        for instance in read_chronos_jobs_for_service(service=service, cluster=cluster, soa_dir=soa_dir):
            if instance.startswith('_'):  # some dependencies are templates, ignore
                continue
            try:
                chronos_configs[(service, instance)] = load_chronos_job_config(
                    service=service,
                    instance=instance,
                    cluster=cluster,
                    soa_dir=soa_dir,
                )
            except NoDeploymentsAvailable:
                pass

    adjacency_list = defaultdict(set)  # List of adjacency used by dfs algorithm
    for instance, config in chronos_configs.items():
        for parent in config.get_parents():
            parent = decompose_job_id(paasta_to_chronos_job_name(parent))
            # Map the graph a undirected graph to simplify identification of connected components
            adjacency_list[parent].add(instance)
            adjacency_list[instance].add(parent)

    def cached_dfs(known_dfs_results, node, neighbours_mapping, ignore_cycles=False):
        """
        Cached version of DFS algorithm (tuned for extraction of connected components).

        :param known_dfs_results: previous results of DFS
        :type known_dfs_results: list
        :param node: starting node for DFS execution
        :param neighbours_mapping: graph adjacency list
        :type neighbours_mapping: dict
        :param ignore_cycles: raise an exception if a cycle is identified in the graph
        :type ignore_cycles: bool

        :return: set of nodes part of the same connected component of `node`
        :type: set
        """
        for known_dfs_result in known_dfs_results:
            if node in known_dfs_result:
                return known_dfs_result

        visited_nodes = set(dfs(
            node=node,
            neighbours_mapping=neighbours_mapping,
            ignore_cycles=ignore_cycles,
        ))
        known_dfs_results.append(visited_nodes)

        return visited_nodes

    # Build connected components
    known_connected_components = []
    for instance in chronos_configs:
        cached_dfs(  # Using cached version of DFS to avoid all algorithm execution if the result it already know
            known_dfs_results=known_connected_components,
            node=instance,
            ignore_cycles=True,  # Operating on an undirected graph to simplify identification of connected components
            neighbours_mapping=adjacency_list,
        )

    connected_components = {}
    for connected_component in known_connected_components:
        for node in connected_component:
            connected_components[node] = connected_component

    return connected_components, chronos_configs


def get_related_jobs_configs(cluster, service, instance, soa_dir=DEFAULT_SOA_DIR, use_cache=True):
    """
    Extract chronos configurations of all Chronos jobs related to the (cluster, service, instance)

    :return: job-config mapping. Job identifier is a tuple (service, instance)
    """
    kwargs = {}
    if not use_cache:
        kwargs['ttl'] = -1  # Disable usage of time_cached value

    connected_components, chronos_configs = _get_related_jobs_and_configs(cluster, soa_dir=soa_dir, **kwargs)
    related_jobs = connected_components[(service, instance)]
    return {
        job: config
        for job, config in chronos_configs.items()
        if job in related_jobs
    }


def topological_sort_related_jobs(cluster, service, instance, soa_dir=DEFAULT_SOA_DIR):
    """
    Extract the topological order of all the jobs related to the specified instance.
    :return: sorted list of related jobs.
            The list is ordered such that job with index `i` could be executed in respect of its dependencies
            if all jobs with index smaller than `i` have terminated.
    """
    def _sort(nodes):
        adjacency_list = {
            node: [
                decompose_job_id(paasta_to_chronos_job_name(parent_job))
                for parent_job in chronos_configs[node].get_parents()
            ]
            for node in related_jobs
        }

        order, enter, node_status = list(), {n for n in nodes}, {}

        while enter:
            order.extend(
                dfs(
                    node=enter.pop(),
                    neighbours_mapping=adjacency_list,
                    node_status=node_status,
                    ignore_cycles=False,
                ),
            )

        return order

    connected_components, chronos_configs = _get_related_jobs_and_configs(cluster, soa_dir=soa_dir)
    related_jobs = connected_components[(service, instance)]

    return _sort(
        nodes=related_jobs,
    )



import time
from typing import Callable
from typing import Dict
from typing import List

from signalfx import SignalFx
from signalfx.signalflow.messages import DataMessage
from signalfx.signalflow.messages import MetadataMessage


def _convert_sfx_timestamp(ts: int) -> float:
    '''SignalFx uses millisecond int timestamps, we want floating point seconds'''
    return float(ts) / 1000


def tail_signalfx(query: str, lookback_seconds: float, callback: Callable, sfx_api_token: str) -> None:
    if lookback_seconds > 0:
        start_timestamp_milliseconds = (time.time() - lookback_seconds) * 1000
    else:
        start_timestamp_milliseconds = None

    with SignalFx().signalflow(sfx_api_token) as flow:
        tsid_metadata_map = {}

        computation = flow.execute(
            program=query,
            start=start_timestamp_milliseconds,
        )
        for msg in computation.stream():
            if isinstance(msg, MetadataMessage):
                # The MetadataMessage occurs before the associated DataMessage.
                # Extract out the important properties since a lot of it is not important.
                tsid, props = msg.tsid, msg.properties

                # in SignalFX, query looks like ... .publish('<stream_label>')
                stream_label: str = props.get('sf_streamLabel', tsid)

                # sf_key seems to have the list of dimensions with a couple extra stuff
                # The SignalFX query should minimally specify the 'service' dimension.
                # .sum(by=['dim1','dim2'])
                sf_key: List[str] = props.get('sf_key', [])

                dimensions: Dict[str, str] = {dim: props[dim] for dim in sf_key}
                tsid_metadata_map[tsid] = {'stream_label': stream_label, 'dimensions': dimensions}
            elif isinstance(msg, DataMessage):
                for tsid, datapoint in msg.data.items():
                    props = tsid_metadata_map[tsid]
                    callback(props, datapoint, _convert_sfx_timestamp(msg.logical_timestamp_ms))



import abc
import json
import logging
import time
import traceback
from multiprocessing import Process
from multiprocessing import Queue
from queue import Empty
from threading import Thread
from typing import Dict
from typing import List
from typing import Optional

import requests
import transitions
from slackclient import SlackClient

from paasta_tools.automatic_rollbacks.state_machine import DeploymentProcess


try:
    from scribereader import scribereader
except ImportError:
    scribereader = None

SLACK_WEBHOOK_STREAM = 'stream_slack_incoming_webhook'
SCRIBE_ENV = 'uswest1-prod'
log = logging.getLogger(__name__)


class ButtonPress():
    def __init__(self, event):
        self.event = event
        self.username = event["user"]["username"]
        self.response_url = event["response_url"]
        # TODO: Handle multiple actions?
        self.action = event["actions"][0]["value"]
        self.thread_ts = event["container"].get("thread_ts", None)
        self.channel = event["channel"]["name"]

    def __repr__(self):
        return self.event

    def update(self, blocks):
        # Implements responding to button presses
        # https://api.slack.com/messaging/interactivity/enabling#responding-to-interactions
        # But isn't the api_call method per-se
        # https://github.com/slackapi/python-slackclient/issues/270
        requests.post(self.response_url, json={"blocks": blocks})


def event_to_buttonpress(event):
    return ButtonPress(event=event)


def parse_webhook_event_json(line):
    event = json.loads(line)
    log.debug(event)
    return event


def is_relevant_event(event):
    # TODO: Implement filtering
    return True


def get_slack_events():
    if scribereader is None:
        logging.error("Scribereader unavailable. Not tailing slack events.")
        return

    def scribe_tail(queue):
        host_and_port = scribereader.get_env_scribe_host(SCRIBE_ENV, True)
        host = host_and_port['host']
        port = host_and_port['port']
        tailer = scribereader.get_stream_tailer(SLACK_WEBHOOK_STREAM, host, port)
        for line in tailer:
            queue.put(line)

    # Tailing scribe is not thread-safe, therefore we must use a Multiprocess-Queue-based
    # approach, with paasta logs as prior art.
    queue = Queue()
    kw = {'queue': queue}
    process = Process(target=scribe_tail, daemon=True, kwargs=kw)
    process.start()
    while True:
        try:
            line = queue.get(block=True, timeout=0.1)
            event = parse_webhook_event_json(line)
            if is_relevant_event(event):
                yield event
        except Empty:
            pass


class SlackDeploymentProcess(DeploymentProcess, abc.ABC):
    def __init__(self) -> None:
        super().__init__()
        self.human_readable_status = "Initializing..."
        self.slack_client = self.get_slack_client()
        self.last_action = None
        self.send_initial_slack_message()

        slack_thread = Thread(target=self.listen_for_slack_events, args=(), daemon=True)
        slack_thread.start()

        timer_thread = Thread(target=self.periodically_update_slack, args=(), daemon=True)
        timer_thread.start()

    @abc.abstractmethod
    def get_slack_client(self) -> SlackClient:
        raise NotImplementedError()

    @abc.abstractmethod
    def get_slack_channel(self) -> str:
        raise NotImplementedError()

    @abc.abstractmethod
    def get_deployment_name(self) -> str:
        raise NotImplementedError()

    @abc.abstractmethod
    def get_progress(self) -> str:
        raise NotImplementedError()

    def get_active_button(self) -> Optional[str]:
        return None

    @abc.abstractmethod
    def get_button_text(self, button, is_active) -> str:
        raise NotImplementedError()

    def get_button_element(self, button, is_active):
        element = {
            "type": "button",
            "text": {
                "type": "plain_text",
                "text": self.get_button_text(button, is_active),
                "emoji": True,
            },
            "value": button,
        }
        if not is_active:
            element["confirm"] = self.get_confirmation_object(button)
        return element

    def get_slack_blocks_for_deployment(self) -> List[Dict]:
        status = getattr(self, 'state', None) or 'Uninitialized'
        deployment_name = self.get_deployment_name()
        message = self.human_readable_status
        progress = self.get_progress()
        last_action = self.last_action

        button_elements = self.get_button_elements()
        blocks = [
            {
                "type": "section",
                "text": {
                    "type": "mrkdwn",
                    "text": f"{deployment_name}",
                },
            },
            {
                "type": "section",
                "text": {
                    "type": "mrkdwn",
                    "text": message,
                },
            },
            {
                "type": "section",
                "text": {
                    "type": "mrkdwn",
                    "text": f"State machine: `{status}`\nProgress: {progress}\nLast operator action: {last_action}",
                },
            },
        ]

        blocks.extend(self.get_extra_blocks_for_deployment())

        if button_elements != []:
            blocks.append({
                "type": "actions",
                "block_id": "deployment_actions",
                "elements": button_elements,
            })
        return blocks

    def get_extra_blocks_for_deployment(self) -> List:
        return []

    def get_button_elements(self):
        elements = []
        active_button = self.get_active_button()
        for button in self.get_available_buttons():
            is_active = button == active_button
            elements.append(
                self.get_button_element(button=button, is_active=is_active),
            )
        return elements

    def get_confirmation_object(self, action):
        return {
            "title": {
                "type": "plain_text",
                "text": "Are you sure?",
            },
            "text": {
                "type": "mrkdwn",
                "text": f"Did you mean to press {action}?",
            },
            "confirm": {
                "type": "plain_text",
                "text": "Yes. Do it!",
            },
            "deny": {
                "type": "plain_text",
                "text": "Stop, I've changed my mind!",
            },
        }

    def get_available_buttons(self) -> List[str]:
        buttons = []

        if self.is_terminal_state(self.state):
            # If we're about to exit, always clear the buttons, since once we exit the buttons will stop working.
            return []

        for trigger in self.machine.get_triggers(self.state):
            suffix = '_button_clicked'
            if trigger.endswith(suffix):
                if all(
                    cond.target == self.machine.resolve_callable(
                        cond.func,
                        transitions.EventData(self.state, self, self.machine, self, args=(), kwargs={}),
                    )()
                    for transition in self.machine.get_transitions(source=self.state, trigger=trigger)
                    for cond in transition.conditions
                ):
                    buttons.append(trigger[:-len(suffix)])

        return buttons

    def update_slack_thread(self, message):
        log.debug(f"Updating slack thread with {message}")
        resp = self.slack_client.api_call(
            'chat.postMessage',
            channel=self.slack_channel,
            text=message,
            thread_ts=self.slack_ts,
        )
        if resp["ok"] is not True:
            log.error("Posting to slack failed: {}".format(resp["error"]))

    def send_initial_slack_message(self):
        blocks = self.get_slack_blocks_for_deployment()
        resp = self.slack_client.api_call('chat.postMessage', blocks=blocks, channel=self.slack_channel)
        self.slack_ts = resp['message']['ts'] if resp and resp['ok'] else None
        self.slack_channel_id = resp['channel']
        if resp["ok"] is not True:
            log.error("Posting to slack failed: {}".format(resp["error"]))

    def update_slack(self):
        blocks = self.get_slack_blocks_for_deployment()
        resp = self.slack_client.api_call(
            "chat.update",
            channel=self.slack_channel_id,
            blocks=blocks,
            ts=self.slack_ts,
        )
        if resp["ok"] is not True:
            log.error("Posting to slack failed: {}".format(resp["error"]))

    def update_slack_status(self, message):
        self.human_readable_status = message
        self.update_slack()

    def periodically_update_slack(self):
        while self.state not in self.status_code_by_state():
            self.update_slack()
            time.sleep(20)

    def is_relevant_buttonpress(self, buttonpress):
        return self.slack_ts == buttonpress.thread_ts

    def listen_for_slack_events(self):
        log.debug("Listening for slack events...")
        for event in get_slack_events():
            log.debug(f"Got slack event: {event}")
            buttonpress = event_to_buttonpress(event)
            if self.is_relevant_buttonpress(buttonpress):
                self.update_slack_thread(f"{buttonpress.username} pressed {buttonpress.action}")
                self.last_action = buttonpress.action

                try:
                    self.trigger(f"{buttonpress.action}_button_clicked")
                except (transitions.core.MachineError, AttributeError):
                    self.update_slack_thread(f"Error: {traceback.format_exc()}")
            else:
                log.debug("But it was not relevant to this instance of mark-for-deployment")

    def notify_users(self, message):
        self.update_slack_thread(message)



#!/usr/bin/env python3.6
import itertools
import json
import sys
from collections import defaultdict
from datetime import datetime

from paasta_tools.utils import paasta_print


def get_datetime_from_ts(ts):
    tformat = '%Y-%m-%dT%H:%M:%S.%f'
    return datetime.strptime(ts, tformat)


def get_deploy_durations_from_file(filename):
    """
    filename: path to a file to be parsed for datetime data
    The expected input is a paasta service log for the deploy events
    The way I've been fetching them is by running 'internal logreader command' | grep deploy | grep event > filename
    """
    file_object = open(filename, 'r')
    data = sorted(
        [json.loads(line.rstrip('\n')) for line in file_object],
        key=lambda x: get_datetime_from_ts(x['timestamp']),
    )

    timedeltas = defaultdict(list)
    last_time = dict()
    instance_bitvector = defaultdict(bool)  # defaults to False

    for datum in data:
        time = get_datetime_from_ts(datum['timestamp'])
        instance = datum['instance']
        if 'in progress' in datum['message'] and not instance_bitvector[instance]:
            instance_bitvector[instance] = True
            last_time[instance] = time
        elif 'finishing' in datum['message']:
            instance_bitvector[instance] = False
            timedeltas[instance].append(time - last_time[instance])

    return timedeltas


def display_bounce_info(timedeltas):
    """
    timedeltas: iterable of timedelta objects
    """
    std = list(sorted(timedeltas))
    paasta_print("Median time to bounce: {} seconds".format(std[len(std) / 2]))
    paasta_print("10% time to bounce: {}".format(std[len(std) / 10]))
    paasta_print("90% time to bounce: {}".format(std[len(std) * 9 / 10]))


def main(filenames):
    for filename in filenames:
        paasta_print(filename)
        paasta_print("=========================")
        timedeltas = get_deploy_durations_from_file(filename)
        for instance, tdlist in timedeltas.items():
            if timedeltas:
                paasta_print("Instance: %s" % instance)
                display_bounce_info(tdlist)
        paasta_print("Overall:")
        display_bounce_info(itertools.chain.from_iterable(timedeltas.values()))
        paasta_print("=========================")


if __name__ == '__main__':
    main(filenames=sys.argv[1:])



#!/usr/bin/env python
# Copyright 2015-2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Client interface for the Paasta rest api.
"""
import json
import logging
import os
from typing import Any
from urllib.parse import urlparse

from bravado.client import SwaggerClient
from bravado.requests_client import RequestsClient

import paasta_tools.api
from paasta_tools.utils import load_system_paasta_config
from paasta_tools.utils import SystemPaastaConfig


log = logging.getLogger(__name__)


def get_paasta_api_client(
    cluster: str = None,
    system_paasta_config: SystemPaastaConfig = None,
    http_res: bool = False,
) -> Any:
    if not system_paasta_config:
        system_paasta_config = load_system_paasta_config()

    if not cluster:
        cluster = system_paasta_config.get_cluster()

    api_endpoints = system_paasta_config.get_api_endpoints()
    if cluster not in api_endpoints:
        log.error('Cluster %s not in paasta-api endpoints config', cluster)
        return None

    url = str(api_endpoints[cluster])
    parsed = urlparse(url)
    if not parsed:
        log.error('Unsupported paasta-api url %s', url)
        return None
    api_server = parsed.netloc

    # Get swagger spec from file system instead of the api server
    paasta_api_path = os.path.dirname(paasta_tools.api.__file__)
    swagger_file = os.path.join(paasta_api_path, 'api_docs/swagger.json')
    if not os.path.isfile(swagger_file):
        log.error('paasta-api swagger spec %s does not exist', swagger_file)
        return None

    with open(swagger_file) as f:
        spec_dict = json.load(f)
    # replace localhost in swagger.json with actual api server
    spec_dict['host'] = api_server
    spec_dict['schemes'] = [parsed.scheme]

    # sometimes we want the status code
    requests_client = PaastaRequestsClient(
        scheme=parsed.scheme,
        cluster=cluster,
        system_paasta_config=system_paasta_config,
    )
    if http_res:
        config = {'also_return_response': True}
        return SwaggerClient.from_spec(spec_dict=spec_dict, config=config, http_client=requests_client)
    else:
        return SwaggerClient.from_spec(spec_dict=spec_dict, http_client=requests_client)


class PaastaRequestsClient(RequestsClient):
    def __init__(
        self,
        scheme: str,
        cluster: str,
        system_paasta_config: SystemPaastaConfig,
    ) -> None:
        if scheme == 'https':
            if system_paasta_config.get_enable_client_cert_auth():
                paasta_dir = os.path.expanduser('~/.paasta/pki')
                ssl_cert = (f'{paasta_dir}/{cluster}.crt', f'{paasta_dir}/{cluster}.key')
            else:
                ssl_cert = None
            super().__init__(ssl_verify=True, ssl_cert=ssl_cert)
            ca_cert_path = system_paasta_config.get_api_ca_certificates().get(
                cluster,
                system_paasta_config.get_default_api_ca_certificate(),
            )
            self.session.verify = ca_cert_path
        else:
            super().__init__()



#!/usr/bin/env python
# Copyright 2015-2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import asyncio
import logging
import math
import os
import time
from abc import abstractmethod
from collections import defaultdict
from datetime import datetime
from datetime import timedelta
from datetime import timezone
from math import ceil
from math import floor
from typing import Any
from typing import DefaultDict
from typing import Dict
from typing import List
from typing import Mapping
from typing import NamedTuple
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import Type
from typing import Union

import a_sync
import boto3
from botocore.exceptions import ClientError
from mypy_extensions import TypedDict
from requests.exceptions import HTTPError

from paasta_tools.autoscaling import ec2_fitness
from paasta_tools.autoscaling import load_boost
from paasta_tools.mesos.master import MesosState
from paasta_tools.mesos_maintenance import drain
from paasta_tools.mesos_maintenance import undrain
from paasta_tools.mesos_tools import get_mesos_master
from paasta_tools.mesos_tools import get_mesos_task_count_by_slave
from paasta_tools.mesos_tools import slave_pid_to_ip
from paasta_tools.mesos_tools import SlaveTaskCount
from paasta_tools.metrics.metastatus_lib import get_resource_utilization_by_grouping
from paasta_tools.metrics.metastatus_lib import ResourceInfo
from paasta_tools.metrics.metrics_lib import get_metrics_interface
from paasta_tools.paasta_maintenance import is_safe_to_kill
from paasta_tools.utils import ClusterAutoscalingResource
from paasta_tools.utils import IdToClusterAutoscalingResourcesDict
from paasta_tools.utils import load_system_paasta_config
from paasta_tools.utils import PoolToResourcePoolSettingsDict
from paasta_tools.utils import ResourcePoolSettings
from paasta_tools.utils import SystemPaastaConfig
from paasta_tools.utils import Timeout
from paasta_tools.utils import TimeoutError


class AutoscalingInfo(NamedTuple):
    resource_id: str
    pool: str
    state: str
    current: str
    target: str
    min_capacity: str
    max_capacity: str
    instances: str


class InstanceDescription(TypedDict):
    InstanceId: str
    PrivateIpAddress: str


class LaunchSpecification(TypedDict):
    InstanceType: str
    WeightedCapacity: float


class SpotFleetRequestConfig(TypedDict):
    FulfilledCapacity: int
    TargetCapacity: int
    LaunchSpecifications: Sequence[LaunchSpecification]


class SpotFleetRequestConfig0(TypedDict):
    SpotFleetRequestState: str
    CreateTime: datetime
    SpotFleetRequestConfig: SpotFleetRequestConfig


CLUSTER_METRICS_PROVIDER_KEY = 'cluster_metrics_provider'
DEFAULT_TARGET_UTILIZATION = 0.8  # decimal fraction
DEFAULT_DRAIN_TIMEOUT = 600  # seconds

AWS_SPOT_MODIFY_TIMEOUT = 30
MISSING_SLAVE_PANIC_THRESHOLD = .3

log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())


class Timer:
    def __init__(self, timeout: int) -> None:
        self.timeout = timedelta(seconds=timeout)
        self.start()

    def start(self) -> None:
        self.last_start = datetime.now()

    def ready(self) -> bool:
        return datetime.now() > self.last_start + self.timeout

    def left(self) -> timedelta:
        return self.last_start + self.timeout - datetime.now()


class ClusterAutoscaler:

    def __init__(
        self,
        resource: ClusterAutoscalingResource,
        pool_settings: ResourcePoolSettings,
        config_folder: str,
        dry_run: bool,
        utilization_error: float,
        max_increase: float,
        max_decrease: float,
        log_level: str = None,
        draining_enabled: bool = True,
        enable_metrics: bool = False,
        enable_maintenance_reservation: bool = True,
    ) -> None:
        self.resource = resource
        self.pool_settings = pool_settings
        self.config_folder = config_folder
        self.dry_run = dry_run
        self.utilization_error = utilization_error
        self.ideal_capacity: Optional[int] = None
        self.draining_enabled = draining_enabled
        self.max_increase = max_increase
        self.max_decrease = max_decrease
        if log_level is not None:
            self.log.setLevel(log_level)
        self.instances: Sequence[InstanceDescription] = []
        self.sfr: Optional[SpotFleetRequestConfig0] = None
        self.enable_metrics = enable_metrics
        self.enable_maintenance_reservation = enable_maintenance_reservation

        self.log.info('Initialized with utilization error %s' % self.utilization_error)
        config = load_system_paasta_config()
        self.slave_newness_threshold = config.get_monitoring_config().get('check_registered_slave_threshold')
        self.setup_metrics(config)

    @property
    def log(self) -> logging.Logger:
        resource_id = self.resource.get("id", "unknown")
        name = '.'.join([__name__, self.__class__.__name__, resource_id])
        return logging.getLogger(name)

    def setup_metrics(self, config: SystemPaastaConfig) -> None:
        if not self.enable_metrics:
            return None
        dims = {
            'paasta_cluster': config.get_cluster(),
            'region': self.resource.get('region', 'unknown'),
            'pool': self.resource.get('pool', 'unknown'),
            'resource_id': self.resource.get('id', 'unknown'),
            'resource_type': self.__class__.__name__,
        }
        self.metrics = get_metrics_interface('paasta.cluster_autoscaler')
        self.target_gauge = self.metrics.create_gauge('target_capacity', **dims)
        self.current_gauge = self.metrics.create_gauge('current_capacity', **dims)
        self.ideal_gauge = self.metrics.create_gauge('ideal_capacity', **dims)
        self.max_gauge = self.metrics.create_gauge('max_capacity', **dims)
        self.min_gauge = self.metrics.create_gauge('min_capacity', **dims)
        self.mesos_error_gauge = self.metrics.create_gauge('mesos_error', **dims)
        self.aws_instances_gauge = self.metrics.create_gauge('aws_instances', **dims)
        self.mesos_slaves_gauge = self.metrics.create_gauge('mesos_slaves', **dims)

    def emit_metrics(
        self,
        current_capacity: float,
        target_capacity: float,
        mesos_slave_count: int,
    ) -> None:
        if not self.enable_metrics:
            return None
        self.current_gauge.set(current_capacity)
        self.target_gauge.set(target_capacity)
        self.ideal_gauge.set(self.ideal_capacity)
        self.min_gauge.set(self.resource['min_capacity'])
        self.max_gauge.set(self.resource['max_capacity'])
        self.mesos_error_gauge.set(self.utilization_error)
        self.aws_instances_gauge.set(len(self.instances))
        self.mesos_slaves_gauge.set(mesos_slave_count)

    def set_capacity(self, capacity: float) -> Optional[Any]:
        pass

    def get_instance_type_weights(self) -> Optional[Mapping[str, float]]:
        pass

    def is_resource_cancelled(self) -> bool:
        raise NotImplementedError()

    def metrics_provider(self, mesos_state: MesosState) -> Tuple[float, int]:
        raise NotImplementedError()

    def describe_instances(
        self,
        instance_ids: Sequence[str],
        region: Optional[str] = None,
        instance_filters: Optional[Sequence[Mapping]] = None,
    ) -> Optional[Sequence[InstanceDescription]]:
        """This wraps ec2.describe_instances and catches instance not
        found errors. It returns a list of instance description
        dictionaries.  Optionally, a filter can be passed through to
        the ec2 call

        :param instance_ids: a list of instance ids, [] means all
        :param instance_filters: a list of ec2 filters
        :param region to connect to ec2
        :returns: a list of instance description dictionaries"""
        if not instance_filters:
            instance_filters = []
        ec2_client = boto3.client('ec2', region_name=region)
        try:
            instance_descriptions = ec2_client.describe_instances(InstanceIds=instance_ids, Filters=instance_filters)
        except ClientError as e:
            if e.response['Error']['Code'] == 'InvalidInstanceID.NotFound':
                self.log.warn(f'Cannot find one or more instance from IDs {instance_ids}')
                return None
            else:
                raise
        instance_reservations = [reservation['Instances'] for reservation in instance_descriptions['Reservations']]
        instances = [instance for reservation in instance_reservations for instance in reservation]
        return instances

    def is_new_autoscaling_resource(self) -> bool:
        raise NotImplementedError()

    def describe_instance_status(
        self,
        instance_ids: Sequence[str],
        region: Optional[str] = None,
        instance_filters: Optional[Sequence[Mapping]] = None,
    ) -> Optional[Mapping]:
        """This wraps ec2.describe_instance_status and catches instance not
        found errors. It returns a list of instance description
        dictionaries.  Optionally, a filter can be passed through to
        the ec2 call

        :param instance_ids: a list of instance ids, [] means all
        :param instance_filters: a list of ec2 filters
        :param region to connect to ec2
        :returns: a list of instance description dictionaries"""
        if not instance_filters:
            instance_filters = []
        ec2_client = boto3.client('ec2', region_name=region)
        try:
            instance_descriptions = ec2_client.describe_instance_status(
                InstanceIds=instance_ids,
                Filters=instance_filters,
            )
        except ClientError as e:
            if e.response['Error']['Code'] == 'InvalidInstanceID.NotFound':
                self.log.warn(f'Cannot find one or more instance from IDs {instance_ids}')
                return None
            else:
                raise
        return instance_descriptions

    def get_instance_ips(
        self,
        instances: Sequence[InstanceDescription],
        region: Optional[str] = None,
    ) -> Sequence[str]:
        instance_descriptions = self.describe_instances(
            [instance['InstanceId'] for instance in instances],
            region=region,
        )
        instance_ips = []
        for instance in instance_descriptions:
            try:
                instance_ips.append(instance['PrivateIpAddress'])
            except KeyError:
                self.log.warning("Instance {} does not have an IP. This normally means it has been"
                                 " terminated".format(instance['InstanceId']))
        return instance_ips

    def cleanup_cancelled_config(
        self,
        resource_id: str,
        config_folder: str,
        dry_run: bool = False,
    ) -> None:
        file_name = f"{resource_id}.json"
        configs_to_delete = [os.path.join(walk[0], file_name)
                             for walk in os.walk(config_folder) if file_name in walk[2]]
        if not configs_to_delete:
            self.log.info(f"Resource config for {resource_id} not found")
            return
        if not dry_run:
            os.remove(configs_to_delete[0])
        self.log.info("Deleted Resource config {}".format(configs_to_delete[0]))

    def get_aws_slaves(self, mesos_state: MesosState) -> Mapping[str, Mapping]:
        instance_ips = self.get_instance_ips(self.instances, region=self.resource['region'])
        slaves = {
            slave['id']: slave for slave in mesos_state.get('slaves', [])
            if slave_pid_to_ip(slave['pid']) in instance_ips and
            slave['attributes'].get('pool', 'default') == self.resource['pool']
        }
        return slaves

    def check_expected_slaves(
        self,
        slaves: Mapping[str, Mapping],
        expected_instances: Optional[int],
    ) -> None:
        current_instances = len(slaves)
        if current_instances == 0:
            error_message = ("No instances are active, not scaling until the instances are attached to mesos")
            raise ClusterAutoscalingError(error_message)
        if not expected_instances:
            return None

        self.log.info("Found %.2f%% slaves registered in mesos for this resource (%d/%d)" % (
            float(float(current_instances) / float(expected_instances)) * 100,
            current_instances,
            expected_instances,
        ))
        if float(current_instances) / expected_instances < (1.00 - MISSING_SLAVE_PANIC_THRESHOLD):
            warning_message = (
                "We currently have %d instances active in mesos out of a desired %d.\n"
            ) % (
                current_instances, expected_instances,
            )

            if self.sfr and self.sfr['SpotFleetRequestState'] == 'cancelled_running':
                warning_message += "But this is an sfr in cancelled_running state, so continuing anyways"
            elif self.utilization_error < 0:
                warning_message += (
                    "Continuing because we are scaling down and we don't need any "
                    "of the unregistered instances."
                )
            else:
                warning_message += (
                    "This may be a sign that new instances are not healthy for some reason. "
                    "Continuing to scale."
                )

            self.log.warn(warning_message)

    async def can_kill(
        self,
        hostname: str,
        should_drain: bool,
        dry_run: bool,
        timer: Timer,
    ) -> bool:
        if dry_run:
            return True
        if timer.ready():
            self.log.warning("Timer expired before slave ready to kill, proceeding to terminate anyways")
            timer.start()
            raise TimeoutError
        if not should_drain:
            self.log.info("Not draining, waiting %s longer before killing" % timer.left())
            return False
        if await a_sync.run(is_safe_to_kill, hostname):
            self.log.info("Slave {} is ready to kill, with {} left on timer".format(hostname, timer.left()))
            timer.start()
            return True
        return False

    async def wait_and_terminate(
        self,
        slave: 'PaastaAwsSlave',
        drain_timeout: int,
        dry_run: bool,
        timer: Timer,
        region: Optional[str] = None,
        should_drain: bool = True,
    ) -> None:
        """Waits for slave to be drained and then terminate

        :param slave: dict of slave to kill
        :param drain_timeout: how long to wait before terminating
            even if not drained
        :param region to connect to ec2
        :param dry_run: Don't drain or make changes to spot fleet if True
        :param should_drain: whether we should drain hosts before waiting to stop them
        :param timer: a Timer object to keep terminates happening once every n seconds across co-routines
        """
        self.log.info("Starting TERMINATING: {} (Hostname = {}, IP = {})".format(
            slave.instance_id,
            slave.hostname,
            slave.ip,
        ))
        try:
            # This loop should always finish because the maintenance window should trigger is_ready_to_kill
            # being true. Just in case though we set a timeout (from timer) and terminate anyway
            while True:
                instance_id = slave.instance_id
                if not instance_id:
                    self.log.warning(
                        f"Didn't find instance ID for slave: {slave.pid}. Skipping terminating",
                    )
                    break
                # Check if no tasks are running or we have reached the maintenance window
                if await self.can_kill(slave.hostname, should_drain, dry_run, timer):
                    self.log.info("TERMINATING: {} (Hostname = {}, IP = {})".format(
                        instance_id,
                        slave.hostname,
                        slave.ip,
                    ))
                    self.terminate_instances([instance_id])
                    break
                else:
                    self.log.info(f"Instance {instance_id}: NOT ready to kill")
                self.log.debug("Waiting 5 seconds and then checking again")
                await asyncio.sleep(5)
        except TimeoutError:
            self.log.error("Timed out after {} waiting to drain {}, now terminating anyway".format(
                timer.timeout,
                slave.pid,
            ))
            self.terminate_instances([instance_id])

    def terminate_instances(self, instance_ids: Sequence[str]) -> None:
        ec2_client = boto3.client('ec2', region_name=self.resource['region'])
        try:
            ec2_client.terminate_instances(
                InstanceIds=instance_ids,
                DryRun=self.dry_run,
            )
        except ClientError as e:
            if e.response['Error'].get('Code') == 'DryRunOperation':
                pass
            else:
                raise

    async def scale_resource(
        self,
        current_capacity: float,
        target_capacity: int,
    ) -> None:
        """Scales an AWS resource based on current and target capacity
        If scaling up we just set target capacity and let AWS take care of the rest
        If scaling down we pick the slaves we'd prefer to kill, put them in maintenance
        mode and drain them (via paasta_maintenance and setup_marathon_jobs). We then kill
        them once they are running 0 tasks or once a timeout is reached

        :param current_capacity: integer current resource capacity
        :param target_capacity: target resource capacity
        """
        target_capacity = int(target_capacity)
        delta = target_capacity - current_capacity
        if delta == 0:
            self.log.info(f"Already at target capacity: {target_capacity}")
            return
        elif delta > 0:
            self.log.info(f"Increasing resource capacity to: {target_capacity}")
            self.set_capacity(target_capacity)
            return
        elif delta < 0:
            mesos_state = await get_mesos_master().state_summary()
            slaves_list = await get_mesos_task_count_by_slave(mesos_state, pool=self.resource['pool'])
            filtered_slaves = self.filter_aws_slaves(slaves_list)
            killable_capacity = round(sum([slave.instance_weight for slave in filtered_slaves]), 2)
            amount_to_decrease = round(delta * -1, 2)
            if amount_to_decrease > killable_capacity:
                instance_ids_in_mesos = {
                    slave.instance_id for slave in filtered_slaves
                }
                instance_ids_not_in_mesos = [
                    instance['InstanceId'] for instance in self.instances
                    if instance['InstanceId'] not in instance_ids_in_mesos
                ]
                self.log.warning(
                    "Didn't find enough candidates to kill. This may mean that "
                    "some instances have not yet joined the cluster. Since we "
                    "are scaling down, we will kill the instances in this "
                    "pool that have not joined.\n"
                    "Desired decrease in capacity: %s, killable capacity: %s\n"
                    "Setting capacity to %s and killing instances %s before "
                    "proceeding scale down" % (
                        amount_to_decrease,
                        killable_capacity,
                        killable_capacity,
                        instance_ids_not_in_mesos,
                    ),
                )
                self.set_capacity(killable_capacity)
                self.terminate_instances(instance_ids_not_in_mesos)

            await self.downscale_aws_resource(
                filtered_slaves=filtered_slaves,
                current_capacity=current_capacity,
                target_capacity=target_capacity,
            )

    def should_drain(self, slave_to_kill: 'PaastaAwsSlave') -> bool:
        if not self.draining_enabled:
            return False
        if self.dry_run:
            return False
        if slave_to_kill.instance_status['SystemStatus']['Status'] != 'ok':
            return False
        if slave_to_kill.instance_status['InstanceStatus']['Status'] != 'ok':
            return False
        return True

    async def gracefully_terminate_slave(
        self,
        slave_to_kill: 'PaastaAwsSlave',
        capacity_diff: int,
        timer: Timer,
    ) -> None:
        """
        Since this is async, it can be suspended at an `await` call.  Because of this, we need to re-calculate
        the capacity each time we call `set_capacity` (as another coroutine could have set the capacity while
        this one was suspended).  `set_capacity` stores the currently set capacity in the object, and then
        this function re-calculates that from the capacity_diff each time we call `set_capacity`
        """
        drain_timeout = self.pool_settings.get('drain_timeout', DEFAULT_DRAIN_TIMEOUT)
        # The start time of the maintenance window is the point at which
        # we giveup waiting for the instance to drain and mark it for termination anyway
        start = int(time.time() + drain_timeout) * 1000000000  # nanoseconds
        # Set the duration to an hour, this is fairly arbitrary as mesos doesn't actually
        # do anything at the end of the maintenance window.
        duration = 600 * 1000000000  # nanoseconds
        self.log.info(f"Draining {slave_to_kill.pid}")
        should_drain = self.should_drain(slave_to_kill)
        if should_drain:
            try:
                drain_host_string = f"{slave_to_kill.hostname}|{slave_to_kill.ip}"
                await a_sync.to_async(drain)(
                    hostnames=[drain_host_string],
                    start=start,
                    duration=duration,
                    reserve_resources=self.enable_maintenance_reservation,
                )
            except HTTPError as e:
                self.log.error("Failed to start drain "
                               "on {}: {}\n Trying next host".format(slave_to_kill.hostname, e))
                raise
        self.log.info("Decreasing resource from {} to: {}".format(self.capacity, self.capacity + capacity_diff))
        # Instance weights can be floats but the target has to be an integer
        # because this is all AWS allows on the API call to set target capacity
        try:
            self.set_capacity(self.capacity + capacity_diff)
        except FailSetResourceCapacity:
            self.log.error("Couldn't update resource capacity, stopping autoscaler")
            self.log.info(f"Undraining {slave_to_kill.pid}")
            if should_drain:
                await a_sync.to_async(undrain)(
                    hostnames=[drain_host_string],
                    unreserve_resources=self.enable_maintenance_reservation,
                )
            raise
        self.log.info("Waiting for instance to drain before we terminate")
        try:
            await self.wait_and_terminate(
                slave=slave_to_kill,
                drain_timeout=drain_timeout,
                dry_run=self.dry_run,
                timer=timer,
                region=self.resource['region'],
                should_drain=should_drain,
            )
            if should_drain:
                await a_sync.to_async(undrain)(
                    hostnames=[drain_host_string],
                    unreserve_resources=self.enable_maintenance_reservation,
                )
        except ClientError as e:
            self.log.error(f"Failure when terminating: {slave_to_kill.pid}: {e}")
            self.log.error("Setting resource capacity back to {}".format(self.capacity - capacity_diff))
            self.set_capacity(self.capacity - capacity_diff)
            self.log.info(f"Undraining {slave_to_kill.pid}")
            if should_drain:
                await a_sync.to_async(undrain)(
                    hostnames=[drain_host_string],
                    unreserve_resources=self.enable_maintenance_reservation,
                )

    def filter_aws_slaves(self, slaves_list: Sequence[Mapping[str, SlaveTaskCount]]) -> Sequence['PaastaAwsSlave']:
        ips = self.get_instance_ips(self.instances, region=self.resource['region'])
        self.log.debug(f"IPs in AWS resources: {ips}")
        slaves = [
            slave for slave in slaves_list
            if slave_pid_to_ip(slave['task_counts'].slave['pid']) in ips
        ]
        slave_ips = [slave_pid_to_ip(slave['task_counts'].slave['pid']) for slave in slaves]
        instance_type_weights = self.get_instance_type_weights()
        instance_statuses = self.instance_status_for_instance_ids(
            instance_ids=[instance['InstanceId'] for instance in self.instances],
        )
        instance_descriptions = self.instance_descriptions_for_ips(slave_ips)

        paasta_aws_slaves = []
        for slave in slaves:
            slave_ip = slave_pid_to_ip(slave['task_counts'].slave['pid'])
            matching_descriptions = self.filter_instance_description_for_ip(slave_ip, instance_descriptions)
            if matching_descriptions:
                assert len(matching_descriptions) == 1, (
                    "There should be only one instance with the same IP."
                    "Found instances %s with the same ip %s"
                    % (
                        ",".join(
                            [x['InstanceId'] for x in matching_descriptions],
                        ),
                        slave_ip,
                    )
                )
                description = matching_descriptions[0]
                matching_status = self.filter_instance_status_for_instance_id(
                    instance_id=description['InstanceId'],
                    instance_statuses=instance_statuses,
                )
                assert len(matching_status) == 1, "There should be only one InstanceStatus per instance"
            else:
                description = None

            paasta_aws_slaves.append(PaastaAwsSlave(
                slave=slave,
                instance_status=matching_status[0],
                instance_description=description,
                instance_type_weights=instance_type_weights,
            ))

        return paasta_aws_slaves

    def instance_status_for_instance_ids(
        self,
        instance_ids: Sequence[str],
    ) -> Mapping[str, Sequence[Mapping[str, Union[str, Mapping, Sequence]]]]:
        """
        Return a list of instance statuses. Batch the API calls into
        groups of 99, since AWS limit it.
        """
        partitions = [
            instance_ids[partition: partition + 99]
            for partition in
            range(0, len(instance_ids), 99)
        ]
        accumulated: Dict[str, List[Dict[str, Union[str, Dict, List]]]] = {
            'InstanceStatuses': [],
        }
        for subgroup in partitions:
            res = self.describe_instance_status(
                instance_ids=subgroup,
                region=self.resource['region'],
            )
            accumulated['InstanceStatuses'] += res['InstanceStatuses']

        return accumulated

    def instance_descriptions_for_ips(self, ips: Sequence[str]) -> Sequence[InstanceDescription]:
        all_instances: List[InstanceDescription] = []
        for start, stop in zip(range(0, len(ips), 199), range(199, len(ips) + 199, 199)):
            all_instances += self.describe_instances(
                instance_ids=[],
                region=self.resource['region'],
                instance_filters=[
                    {
                        'Name': 'private-ip-address',
                        'Values': ips[start:stop],
                    },
                ],
            )
        return all_instances

    def filter_instance_description_for_ip(
        self,
        ip: str,
        instance_descriptions: Sequence[InstanceDescription],
    ) -> Sequence[InstanceDescription]:
        return [
            i for i in instance_descriptions
            if ip == i['PrivateIpAddress']
        ]

    def filter_instance_status_for_instance_id(
        self,
        instance_id: str,
        instance_statuses: Mapping[str, Sequence[Mapping[str, Union[str, Mapping, Sequence]]]],
    ) -> Sequence[Mapping[str, Union[str, Mapping, Sequence]]]:
        return [
            status for status in instance_statuses['InstanceStatuses']
            if status['InstanceId'] == instance_id
        ]

    async def downscale_aws_resource(
        self,
        filtered_slaves: Sequence['PaastaAwsSlave'],
        current_capacity: float,
        target_capacity: int,
    ) -> None:
        self.log.info("downscale_aws_resource for %s" % filtered_slaves)
        killed_slaves = 0
        terminate_tasks = {}
        self.capacity = current_capacity
        timer = Timer(300)
        while True:
            filtered_sorted_slaves = ec2_fitness.sort_by_ec2_fitness(filtered_slaves)[::-1]
            if len(filtered_sorted_slaves) == 0:
                self.log.info("ALL slaves killed so moving on to next resource!")
                break
            self.log.info("Resource slave kill preference: {}".format([
                slave.hostname
                for slave in filtered_sorted_slaves
            ]))
            slave_to_kill = filtered_sorted_slaves.pop(0)
            instance_capacity = slave_to_kill.instance_weight
            new_capacity = current_capacity - instance_capacity
            if new_capacity < target_capacity:
                self.log.info("Terminating instance {} with weight {} would take us below our target of {},"
                              " so this is as close to our target as we can get".format(
                                  slave_to_kill.instance_id,
                                  slave_to_kill.instance_weight,
                                  target_capacity,
                              ))
                if self.resource['type'] == 'aws_spot_fleet_request' and killed_slaves == 0:
                    self.log.info(
                        "This is a SFR so we must kill at least one slave to prevent the autoscaler "
                        "getting stuck whilst scaling down gradually",
                    )
                    if new_capacity < 1 and self.sfr['SpotFleetRequestState'] == 'active':
                        self.log.info(
                            "Can't set target capacity to less than 1 for SFRs. No further "
                            "action for this SFR",

                        )
                        break
                else:
                    break

            capacity_diff = new_capacity - current_capacity
            self.log.info("Starting async kill for %s" % slave_to_kill.hostname)
            # My understanding is that ensure_future will actually start running the coroutine
            #  (gracefully_terminate_slave), until it hits something that sleeps, then the loop
            #  can continue and we start killing the next slave
            terminate_tasks[slave_to_kill.hostname] = asyncio.ensure_future(
                self.gracefully_terminate_slave(
                    slave_to_kill=slave_to_kill,
                    capacity_diff=capacity_diff,
                    timer=timer,
                ),
            )
            killed_slaves += 1

            current_capacity = new_capacity
            filtered_slaves = filtered_sorted_slaves

        # Now we wait for each task to actually finish...
        for hostname, task in terminate_tasks.items():
            try:
                await task
            except (HTTPError, FailSetResourceCapacity):
                continue

    @property
    @abstractmethod
    def exists(self) -> bool:
        raise NotImplementedError()

    @property
    @abstractmethod
    def current_capacity(self) -> float:
        raise NotImplementedError()


class SpotAutoscaler(ClusterAutoscaler):

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.sfr: SpotFleetRequestConfig0 = self.get_sfr(self.resource['id'], region=self.resource['region'])
        if self.sfr:
            self.instances = self.get_spot_fleet_instances(self.resource['id'], region=self.resource['region'])
            if self.sfr['SpotFleetRequestState'] == 'cancelled_running' and self.utilization_error < 0:
                self.log.warn(
                    'We are in cancelled_running with utilization_error < 0, '
                    'resetting utilization error to -1, min_capacity to 0 in '
                    'order to fully scale down.',
                )
                self.utilization_error = -1
                self.resource['min_capacity'] = 0

            self.ideal_capacity = self.get_ideal_capacity()

    @property
    def exists(self) -> bool:
        return False if not self.sfr or self.sfr['SpotFleetRequestState'] == 'cancelled' else True

    def get_sfr(
        self,
        spotfleet_request_id: str,
        region: Optional[str] = None,
    ) -> Optional[SpotFleetRequestConfig0]:
        ec2_client = boto3.client('ec2', region_name=region)
        try:
            sfrs = ec2_client.describe_spot_fleet_requests(SpotFleetRequestIds=[spotfleet_request_id])
        except ClientError as e:
            if e.response['Error']['Code'] == 'InvalidSpotFleetRequestId.NotFound':
                self.log.warn(f'Cannot find SFR {spotfleet_request_id}')
                return None
            else:
                raise
        ret = sfrs['SpotFleetRequestConfigs'][0]
        return ret

    def is_new_autoscaling_resource(self) -> bool:
        """
        Determines if the spot fleet request was created recently as defined by
        CHECK_REGISTERED_SLAVE_THRESHOLD.
        """
        if not self.sfr:
            self.log.warn('Cannot find SFR {}'.format(self.resource['id']))
            return True

        now = datetime.now(timezone.utc)
        return (now - self.sfr['CreateTime']).total_seconds() < self.slave_newness_threshold

    def get_spot_fleet_instances(
        self,
        spotfleet_request_id: str,
        region: Optional[str] = None,
    ) -> Sequence[InstanceDescription]:
        ec2_client = boto3.client('ec2', region_name=region)
        spot_fleet_instances = ec2_client.describe_spot_fleet_instances(
            SpotFleetRequestId=spotfleet_request_id,
        )['ActiveInstances']
        return spot_fleet_instances

    def metrics_provider(self, mesos_state: MesosState) -> Tuple[float, int]:
        if not self.sfr or self.sfr['SpotFleetRequestState'] == 'cancelled':
            self.log.error("SFR not found or cancelled, removing config file.")
            self.cleanup_cancelled_config(self.resource['id'], self.config_folder, dry_run=self.dry_run)
            return 0, 0
        elif self.sfr['SpotFleetRequestState'] == 'cancelled_running' and len(self.instances) == 0:
            self.log.error(
                "This is a cancelled_running SFR with no instances. This is not "
                "a valid SFR state, and it should eventually be cancelled by AWS. "
                "Removing its config.",
            )
            self.cleanup_cancelled_config(self.resource['id'], self.config_folder, dry_run=self.dry_run)
            return 0, 0
        elif self.sfr['SpotFleetRequestState'] in ['submitted', 'modifying', 'cancelled_terminating']:
            self.log.warning(
                "Not scaling an SFR in state: {} so {}, skipping...".format(
                    self.sfr['SpotFleetRequestState'],
                    self.resource['id'],
                ),
            )
            return 0, 0
        elif self.sfr['SpotFleetRequestState'] == 'cancelled_running' and self.utilization_error > 0:
            self.log.warning("Cannot scale cancelled_running SFR upwards, skipping.")
            return 0, 0
        elif self.sfr['SpotFleetRequestState'] not in ['cancelled_running', 'active']:
            self.log.error("Unexpected SFR state: {} for {}".format(
                self.sfr['SpotFleetRequestState'],
                self.resource['id'],
            ))
            raise ClusterAutoscalingError

        expected_instances = len(self.instances)
        if expected_instances == 0:
            self.log.warning(
                "No instances found in SFR, this shouldn't be possible so we "
                "do nothing",
            )
            return 0, 0
        slaves = self.get_aws_slaves(mesos_state)
        self.check_expected_slaves(slaves, expected_instances)
        current, target = self.get_spot_fleet_delta()

        self.emit_metrics(current, target, mesos_slave_count=len(slaves))
        return current, target

    def is_aws_launching_instances(self) -> bool:
        fulfilled_capacity = self.sfr['SpotFleetRequestConfig']['FulfilledCapacity']
        target_capacity = self.sfr['SpotFleetRequestConfig']['TargetCapacity']
        return target_capacity > fulfilled_capacity

    @property
    def current_capacity(self) -> float:
        return float(self.sfr['SpotFleetRequestConfig']['FulfilledCapacity'])

    def get_ideal_capacity(self) -> int:
        return int(ceil((1 + self.utilization_error) * self.current_capacity))

    def get_spot_fleet_delta(self) -> Tuple[float, int]:
        current_capacity = self.current_capacity
        new_capacity = int(min(
            max(
                self.resource['min_capacity'],
                floor(current_capacity * (1.00 - self.max_decrease)),
                self.ideal_capacity,

                # Can only scale a cancelled_running SFR to 0 instances
                0 if self.sfr['SpotFleetRequestState'] == 'cancelled_running' else 1,
            ),
            ceil(current_capacity * (1.00 + self.max_increase)),
            self.resource['max_capacity'],
        ))
        new_capacity = max(new_capacity, self.resource['min_capacity'])
        self.log.debug("The ideal capacity to scale to is %d instances" % self.ideal_capacity)
        self.log.debug("The capacity we will scale to is %d instances" % new_capacity)
        if self.ideal_capacity > self.resource['max_capacity']:
            self.log.warning(
                "Our ideal capacity (%d) is higher than max_capacity (%d). Consider raising max_capacity!" % (
                    self.ideal_capacity, self.resource['max_capacity'],
                ),
            )
        if self.ideal_capacity < self.resource['min_capacity']:
            self.log.warning(
                "Our ideal capacity (%d) is lower than min_capacity (%d). Consider lowering min_capacity!" % (
                    self.ideal_capacity, self.resource['min_capacity'],
                ),
            )
        return current_capacity, new_capacity

    def set_capacity(self, capacity: float) -> Optional[Any]:
        """ AWS won't modify a request that is already modifying. This
        function ensures we wait a few seconds in case we've just modified
        a SFR"""
        rounded_capacity = int(floor(capacity))
        ec2_client = boto3.client('ec2', region_name=self.resource['region'])
        with Timeout(seconds=AWS_SPOT_MODIFY_TIMEOUT):
            try:
                state = None
                while True:
                    state = self.get_sfr(self.resource['id'], region=self.resource['region'])['SpotFleetRequestState']
                    if state == 'active':
                        break
                    if state == 'cancelled_running':
                        self.log.info(
                            "Not updating target capacity because this is a cancelled SFR, "
                            "we are just draining and killing the instances",
                        )
                        return None
                    self.log.debug("SFR {} in state {}, waiting for state: active".format(self.resource['id'], state))
                    self.log.debug("Sleep 5 seconds")
                    time.sleep(5)
            except TimeoutError:
                self.log.error("Spot fleet {} not in active state so we can't modify it.".format(self.resource['id']))
                raise FailSetResourceCapacity
        if self.dry_run:
            return True
        try:
            ret = ec2_client.modify_spot_fleet_request(
                SpotFleetRequestId=self.resource['id'], TargetCapacity=rounded_capacity,
                ExcessCapacityTerminationPolicy='noTermination',
            )
        except ClientError as e:
            self.log.error(f"Error modifying spot fleet request: {e}")
            raise FailSetResourceCapacity
        self.capacity = capacity
        return ret

    def is_resource_cancelled(self) -> bool:
        if not self.sfr:
            return True
        state = self.sfr['SpotFleetRequestState']
        if state in ['cancelled', 'cancelled_running']:
            return True
        return False

    def get_instance_type_weights(self) -> Mapping[str, float]:
        launch_specifications = self.sfr['SpotFleetRequestConfig']['LaunchSpecifications']
        return {ls['InstanceType']: ls['WeightedCapacity'] for ls in launch_specifications}


class AsgAutoscaler(ClusterAutoscaler):

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.asg = self.get_asg(self.resource['id'], region=self.resource['region'])
        if self.asg:
            self.instances = self.asg['Instances']

            if len(self.instances) == 0:
                self.log.warning(
                    "This ASG has no instances, delta should be 1 to "
                    "launch first instance unless max/min capacity override",
                )
                self.utilization_error = 1

            self.ideal_capacity = self.get_ideal_capacity()

    @property
    def exists(self) -> bool:
        return True if self.asg else False

    def is_new_autoscaling_resource(self) -> bool:
        """
        Determines if an autoscaling group was created recently as defined by
        CHECK_REGISTERED_SLAVE_THRESHOLD.
        """
        if not self.asg:
            self.log.warning("ASG {} not found, removing config file".format(self.resource['id']))
            return True

        now = datetime.now(timezone.utc)
        return (now - self.asg['CreatedTime']).total_seconds() < self.slave_newness_threshold

    def get_asg(self, asg_name: str, region: Optional[str] = None) -> Optional[Mapping[str, Any]]:
        asg_client = boto3.client('autoscaling', region_name=region)
        asgs = asg_client.describe_auto_scaling_groups(AutoScalingGroupNames=[asg_name])
        try:
            return asgs['AutoScalingGroups'][0]
        except IndexError:
            self.log.warning(f"No ASG found in this account with name: {asg_name}")
            return None

    def metrics_provider(self, mesos_state: MesosState) -> Tuple[int, int]:
        if not self.asg:
            self.log.warning("ASG {} not found, removing config file".format(self.resource['id']))
            self.cleanup_cancelled_config(self.resource['id'], self.config_folder, dry_run=self.dry_run)
            return 0, 0
        if self.is_aws_launching_instances():
            self.log.warning(
                "ASG still launching new instances so we won't make any"
                "changes this time.",
            )
            return 0, 0
        expected_instances = len(self.instances)
        slaves = self.get_aws_slaves(mesos_state)
        self.check_expected_slaves(slaves, expected_instances)
        current, target = self.get_asg_delta()

        self.emit_metrics(current, target, mesos_slave_count=len(slaves))
        return current, target

    def is_aws_launching_instances(self) -> bool:
        fulfilled_capacity = len(self.asg['Instances'])
        target_capacity = self.asg['DesiredCapacity']
        return target_capacity > fulfilled_capacity

    @property
    def current_capacity(self) -> int:
        return len(self.asg['Instances'])

    def get_ideal_capacity(self) -> int:
        if self.current_capacity == 0:
            return int(min(
                max(1, self.resource['min_capacity']),
                self.resource['max_capacity'],
            ))
        else:
            return int(ceil((1 + self.utilization_error) * self.current_capacity))

    def get_asg_delta(self) -> Tuple[int, int]:
        current_capacity = self.current_capacity
        if current_capacity == 0:
            new_capacity = self.ideal_capacity
        else:
            new_capacity = int(min(
                max(
                    self.resource['min_capacity'],
                    floor(current_capacity * (1.00 - self.max_decrease)),
                    self.ideal_capacity,
                    0,
                ),
                ceil(current_capacity * (1.00 + self.max_increase)),
                # if max and min set to 0 we still drain gradually
                max(self.resource['max_capacity'], floor(current_capacity * (1.00 - self.max_decrease))),
            ))
        new_capacity = max(new_capacity, self.resource['min_capacity'])
        self.log.debug("The ideal capacity to scale to is %d instances" % self.ideal_capacity)
        self.log.debug("The capacity we will scale to is %d instances" % new_capacity)
        if self.ideal_capacity > self.resource['max_capacity']:
            self.log.warning(
                "Our ideal capacity (%d) is higher than max_capacity (%d). Consider raising max_capacity!" % (
                    self.ideal_capacity, self.resource['max_capacity'],
                ),
            )
        if self.ideal_capacity < self.resource['min_capacity']:
            self.log.warning(
                "Our ideal capacity (%d) is lower than min_capacity (%d). Consider lowering min_capacity!" % (
                    self.ideal_capacity, self.resource['min_capacity'],
                ),
            )
        return current_capacity, new_capacity

    def set_capacity(self, capacity: float) -> Optional[Any]:
        if self.dry_run:
            return True
        asg_client = boto3.client('autoscaling', region_name=self.resource['region'])
        try:
            ret = asg_client.update_auto_scaling_group(
                AutoScalingGroupName=self.resource['id'],
                DesiredCapacity=capacity,
            )
        except ClientError as e:
            self.log.error(f"Error modifying ASG: {e}")
            raise FailSetResourceCapacity
        self.capacity = capacity
        return ret

    def is_resource_cancelled(self) -> bool:
        return not self.asg

    def get_instance_type_weights(self) -> None:
        # None means that all instances will be
        # assumed to have a weight of 1
        return None


class ClusterAutoscalingError(Exception):
    pass


class FailSetResourceCapacity(Exception):
    pass


class PaastaAwsSlave:
    """
    Defines a slave object for use by the autoscaler, containing the mesos slave
    object from mesos state and some properties from AWS
    """

    def __init__(
        self,
        slave: Mapping[str, SlaveTaskCount],
        instance_description: Mapping[str, Any],
        instance_status: Optional[Mapping[str, Any]] = None,
        instance_type_weights: Optional[Mapping] = None,
    ) -> None:
        if instance_status is None:
            self.instance_status: Mapping[str, Any] = {}
        else:
            self.instance_status = instance_status
        self.wrapped_slave = slave
        self.instance_description = instance_description
        self.instance_type_weights = instance_type_weights
        self.task_counts = slave['task_counts']
        self.slave = self.task_counts.slave
        self.ip = slave_pid_to_ip(self.slave['pid'])

    @property
    def instance(self) -> Mapping[str, Any]:
        return self.instance_description

    @property
    def instance_id(self) -> str:
        return self.instance['InstanceId']

    @property
    def hostname(self) -> str:
        return self.slave['hostname']

    @property
    def pid(self) -> str:
        return self.slave['pid']

    @property
    def instance_type(self) -> str:
        return self.instance_description['InstanceType']

    @property
    def instance_weight(self) -> float:
        if self.instance_type_weights:
            return self.instance_type_weights[self.instance_type]
        else:
            return 1


def get_all_utilization_errors(
    autoscaling_resources: IdToClusterAutoscalingResourcesDict,
    all_pool_settings: PoolToResourcePoolSettingsDict,
    mesos_state: MesosState,
    system_config: SystemPaastaConfig,
) -> Mapping[Tuple[str, str], float]:
    errors: Dict[Tuple[str, str], float] = {}
    for identifier, resource in autoscaling_resources.items():
        pool = resource['pool']
        region = resource['region']
        if (region, pool) in errors.keys():
            continue

        target_utilization = all_pool_settings.get(
            pool, {
                'target_utilization': DEFAULT_TARGET_UTILIZATION,
                'drain_timeout': DEFAULT_DRAIN_TIMEOUT,
            },
        ).get(
            'target_utilization', DEFAULT_TARGET_UTILIZATION,
        )
        errors[(region, pool)] = get_mesos_utilization_error(
            mesos_state=mesos_state,
            system_config=system_config,
            region=region,
            pool=pool,
            target_utilization=target_utilization,
        )

    return errors


async def autoscale_local_cluster(
    config_folder: str,
    dry_run: bool = False,
    log_level: str = None,
) -> None:
    log.debug("Sleep 20s to throttle AWS API calls")
    time.sleep(20)
    if dry_run:
        log.info("Running in dry_run mode, no changes should be made")
    system_config = load_system_paasta_config()
    autoscaling_resources = system_config.get_cluster_autoscaling_resources()
    autoscaling_draining_enabled = system_config.get_cluster_autoscaling_draining_enabled()
    all_pool_settings = system_config.get_resource_pool_settings()
    mesos_state = await get_mesos_master().state()
    utilization_errors = get_all_utilization_errors(
        autoscaling_resources=autoscaling_resources,
        all_pool_settings=all_pool_settings,
        mesos_state=mesos_state,
        system_config=system_config,
    )
    log.info("Utilization errors: %s" % utilization_errors)
    autoscaling_scalers: DefaultDict[Tuple[str, str], List[ClusterAutoscaler]] = defaultdict(list)
    for identifier, resource in autoscaling_resources.items():
        pool_settings = all_pool_settings.get(
            resource['pool'], {
                'target_utilization': DEFAULT_TARGET_UTILIZATION,
                'drain_timeout': DEFAULT_DRAIN_TIMEOUT,
            },
        )
        try:
            scaler = get_scaler(resource['type'])(
                resource=resource,
                pool_settings=pool_settings,
                config_folder=config_folder,
                dry_run=dry_run,
                max_increase=system_config.get_cluster_autoscaler_max_increase(),
                max_decrease=system_config.get_cluster_autoscaler_max_decrease(),
                log_level=log_level,
                utilization_error=utilization_errors[(resource['region'], resource['pool'])],
                draining_enabled=autoscaling_draining_enabled,
                enable_metrics=True,
                enable_maintenance_reservation=system_config.get_maintenance_resource_reservation_enabled(),
            )
            autoscaling_scalers[(resource['region'], resource['pool'])].append(scaler)
        except KeyError:
            log.warning("Couldn't find a metric provider for resource of type: {}".format(resource['type']))
            continue
        log.debug("Sleep 3s to throttle AWS API calls")
        time.sleep(3)
    filtered_autoscaling_scalers = filter_scalers(autoscaling_scalers, utilization_errors)
    sorted_autoscaling_scalers = sort_scalers(filtered_autoscaling_scalers)
    await run_parallel_scalers(sorted_autoscaling_scalers, mesos_state)


def sort_scalers(filtered_autoscaling_scalers: Sequence[ClusterAutoscaler]) -> Sequence[ClusterAutoscaler]:
    return sorted(
        filtered_autoscaling_scalers, key=lambda x: x.is_resource_cancelled(), reverse=True,
    )


def filter_scalers(
    autoscaling_scalers: Mapping[Tuple[str, str], Sequence[ClusterAutoscaler]],
    utilization_errors: Mapping[Tuple[str, str], float],
) -> Sequence[ClusterAutoscaler]:
    any_cancelled_in_pool: DefaultDict[Tuple[str, str], bool] = defaultdict(lambda: False)
    for (region, pool), scalers in autoscaling_scalers.items():
        for scaler in scalers:
            if scaler.is_resource_cancelled():
                any_cancelled_in_pool[(region, pool)] = True

    filtered_autoscaling_scalers: List[ClusterAutoscaler] = []

    for (region, pool), scalers in autoscaling_scalers.items():
        if any_cancelled_in_pool[(region, pool)] and utilization_errors[(region, pool)] < 0:
            filtered_autoscaling_scalers += [s for s in scalers if s.is_resource_cancelled()]
            skipped: List[ClusterAutoscaler] = [s for s in scalers if not s.is_resource_cancelled()]
            log.info("There are cancelled resources in pool %s, skipping active resources" % pool)
            log.info("Skipped: %s" % skipped)
        else:
            filtered_autoscaling_scalers += scalers

    return filtered_autoscaling_scalers


async def run_parallel_scalers(
    sorted_autoscaling_scalers: Sequence[ClusterAutoscaler],
    mesos_state: MesosState,
) -> None:
    scaling_tasks = []
    for scaler in sorted_autoscaling_scalers:
        scaling_tasks.append(asyncio.ensure_future(autoscale_cluster_resource(scaler, mesos_state)))
        log.debug("Sleep 3s to throttle AWS API calls")
        await asyncio.sleep(3)
    for task in scaling_tasks:
        if task.cancelled() or task.done():
            continue
        await task


def get_scaler(
    scaler_type: str,
) -> Type[ClusterAutoscaler]:
    scalers = {
        'aws_spot_fleet_request': SpotAutoscaler,
        'aws_autoscaling_group': AsgAutoscaler,
    }
    return scalers[scaler_type]


async def autoscale_cluster_resource(
    scaler: ClusterAutoscaler,
    mesos_state: MesosState,
) -> None:
    log.info("Autoscaling {} in pool, {}".format(scaler.resource['id'], scaler.resource['pool']))
    try:
        current, target = scaler.metrics_provider(mesos_state)
        log.info(f"Target capacity: {target}, Capacity current: {current}")
        await scaler.scale_resource(current, target)
    except ClusterAutoscalingError as e:
        log.error('{}: {}'.format(scaler.resource['id'], e))


def get_instances_from_ip(
    ip: str,
    instance_descriptions: Sequence[InstanceDescription],
) -> Sequence[InstanceDescription]:
    """Filter AWS instance_descriptions based on PrivateIpAddress

    :param ip: private IP of AWS instance.
    :param instance_descriptions: list of AWS instance description dicts.
    :returns: list of instance description dicts"""
    instances = [instance for instance in instance_descriptions if instance['PrivateIpAddress'] == ip]
    return instances


def get_autoscaling_info_for_all_resources(
    mesos_state: MesosState,
) -> Sequence[AutoscalingInfo]:
    system_config = load_system_paasta_config()
    autoscaling_resources = system_config.get_cluster_autoscaling_resources()
    all_pool_settings = system_config.get_resource_pool_settings()
    utilization_errors = get_all_utilization_errors(
        autoscaling_resources=autoscaling_resources,
        all_pool_settings=all_pool_settings,
        mesos_state=mesos_state,
        system_config=system_config,
    )
    vals = [
        autoscaling_info_for_resource(
            resource=resource,
            # The following line is supposed to be
            # ```
            # pool_settings=all_pool_settings[resource['pool']],
            # ```
            # but since
            # 1. we're missing settings for some pools (e.g. `spam` and `kew`)
            # 2. `pool_settings` wasn't actually used to configure autoscalers
            #    as it contained `Mapping[Pool, ResourcePoolSettins]` instead of
            #    `ResourcePoolSettins`, so the only `get()` for `'drain_timeout'`
            #    key always returns the default drain timeout value
            # 3. we're going to replace this autoscaler with Clusterman anyway
            # let just hack it here.
            pool_settings={},  # type: ignore
            mesos_state=mesos_state,
            utilization_errors=utilization_errors,
            max_increase=system_config.get_cluster_autoscaler_max_increase(),
            max_decrease=system_config.get_cluster_autoscaler_max_decrease(),
        )
        for resource in autoscaling_resources.values()
    ]
    return [x for x in vals if x is not None]


def autoscaling_info_for_resource(
    resource: ClusterAutoscalingResource,
    pool_settings: ResourcePoolSettings,
    mesos_state: MesosState,
    utilization_errors: Mapping[Tuple[str, str], float],
    max_increase: float,
    max_decrease: float,
) -> Optional[AutoscalingInfo]:
    scaler_ref = get_scaler(resource['type'])
    scaler = scaler_ref(
        resource=resource,
        pool_settings=pool_settings,
        config_folder=None,
        max_increase=max_increase,
        max_decrease=max_decrease,
        dry_run=True,
        utilization_error=utilization_errors[(resource['region'], resource['pool'])],
    )
    if not scaler.exists:
        log.info("no scaler for resource {}. ignoring".format(resource['id']))
        return None
    target_capacity: Union[int, str]
    try:
        current_capacity, target_capacity = scaler.metrics_provider(mesos_state)
    except ClusterAutoscalingError:
        current_capacity, target_capacity = scaler.current_capacity, "Exception"
    return AutoscalingInfo(
        resource_id=scaler.resource['id'],
        pool=scaler.resource['pool'],
        state="active" if not scaler.is_resource_cancelled() else "cancelled",
        current=str(current_capacity),
        target=str(target_capacity),
        min_capacity=str(scaler.resource['min_capacity']),
        max_capacity=str(scaler.resource['max_capacity']),
        instances=str(len(scaler.instances)),
    )


def get_mesos_utilization_error(
    mesos_state: MesosState,
    system_config: SystemPaastaConfig,
    region: str,
    pool: str,
    target_utilization: float,
) -> float:
    """Return the relative capacity needed to reach the cluster target usage.
    Example: If the current capacity is 10 unit (could be CPU, memory, disk, gpu...)
    If the target usage is 0.8 and the current usage is 9 units. We will return 1.125:
    An 12.5% increase in capacity is required => 9/11.25 = 80% usage
    When the boost feature is enabled, the current_load will be artificially increased
    and stored into boosted_load. If the boost is disabled, boosted_load = current_load
    """
    try:
        region_pool_utilization_dict = get_resource_utilization_by_grouping(
            lambda slave: (
                ('pool', slave['attributes']['pool']),
                ('region', slave['attributes']['datacenter']),
            ),
            mesos_state,
        )[(('pool', pool), ('region', region))]
    except KeyError:
        log.info(
            "Failed to find utilization for region %s, pool %s, returning 0 error" %
            (region, pool),
        )
        return 0

    log.debug(repr(region_pool_utilization_dict))
    usage_percs = []
    for resource in ResourceInfo._fields:
        free = getattr(region_pool_utilization_dict['free'], resource)
        total = getattr(region_pool_utilization_dict['total'], resource)

        if math.isclose(total, 0):
            continue

        current_load = total - free

        # We apply the boost only on the cpu resource.
        if resource == 'cpus' and system_config.get_cluster_boost_enabled():
            boosted_load = load_boost.get_boosted_load(
                zk_boost_path=load_boost.get_zk_cluster_boost_path(region=region, pool=pool),
                current_load=current_load,
            )
        else:
            boosted_load = current_load

        usage_percs.append(boosted_load / float(total))

    if len(usage_percs) == 0:  # If all resource totals are close to 0 for some reason
        return 0

    # We only look at the percentage of utilization. Whichever is the highest (closer or past the setpoint)
    # will be used to determine how much we need to scale
    utilization = max(usage_percs)
    return utilization - target_utilization



#!/usr/bin/env python
# Copyright 2015-2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
PaaSTA service list (instances) etc.
"""
from pyramid.response import Response
from pyramid.view import view_config

from paasta_tools.api import settings
from paasta_tools.api.views.exception import ApiFailure
from paasta_tools.long_running_service_tools import set_instances_for_marathon_service
from paasta_tools.marathon_tools import load_marathon_service_config


@view_config(route_name='service.autoscaler.get', request_method='GET', renderer='json')
def get_autoscaler_count(request):
    service = request.swagger_data.get('service')
    instance = request.swagger_data.get('instance')
    cluster = settings.cluster
    soa_dir = settings.soa_dir
    try:
        service_config = load_marathon_service_config(
            service=service,
            instance=instance,
            cluster=cluster,
            soa_dir=soa_dir,
            load_deployments=False,
        )
    except Exception:
        error_message = f'Unable to load service config for {service}.{instance}'
        raise ApiFailure(error_message, 404)

    response_body = {
        'desired_instances': service_config.get_instances(),
        'calculated_instances': service_config.get_instances(with_limit=False),
    }
    return Response(json_body=response_body, status_code=200)


@view_config(route_name='service.autoscaler.post', request_method='POST', renderer='json')
def update_autoscaler_count(request):
    service = request.swagger_data.get('service')
    instance = request.swagger_data.get('instance')
    desired_instances = request.swagger_data.get('json_body')['desired_instances']
    if not isinstance(desired_instances, int):
        error_message = 'The provided body does not have an integer value for "desired_instances": {}'.format(
            request.swagger_data.get('json_body'),
        )
        raise ApiFailure(error_message, 500)

    try:
        service_config = load_marathon_service_config(
            service=service,
            instance=instance,
            cluster=settings.cluster,
            soa_dir=settings.soa_dir,
            load_deployments=False,
        )
    except Exception:
        error_message = f'Unable to load service config for {service}.{instance}'
        raise ApiFailure(error_message, 404)

    max_instances = service_config.get_max_instances()
    if max_instances is None:
        error_message = f'Autoscaling is not enabled for {service}.{instance}'
        raise ApiFailure(error_message, 404)

    min_instances = service_config.get_min_instances()

    # Dump whatever number from the client to zk. get_instances() will limit
    # readings from zk to [min_instances, max_instances].
    set_instances_for_marathon_service(service=service, instance=instance, instance_count=desired_instances)
    status = 'SUCCESS'
    if desired_instances > max_instances:
        desired_instances = max_instances
        status = 'WARNING desired_instances is greater than max_instances %d' % max_instances
    elif desired_instances < min_instances:
        desired_instances = min_instances
        status = 'WARNING desired_instances is less than min_instances %d' % min_instances

    response_body = {'desired_instances': desired_instances, 'status': status}
    return Response(json_body=response_body, status_code=202)



#!/usr/bin/env python
# Copyright 2015-2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Usage: ./list_kubernetes_service_instances.py [options]

Enumerates all kubernetes instances for services in the soa directory that
are for the current cluster (defined by the kubernetes configuration file).

Outputs (to stdout) a list of service.instance (one per line)
for each instance found in kubernetes-<CLUSTER>.yaml for every folder
in the SOA Configuration directory.

Command line options:

- -d <SOA_DIR>, --soa-dir <SOA_DIR>: Specify a SOA config dir to read from
- -c <CLUSTER>, --cluster <CLUSTER>: Specify which cluster of services to read
"""
import argparse
import sys

from paasta_tools.utils import compose_job_id
from paasta_tools.utils import DEFAULT_SOA_DIR
from paasta_tools.utils import get_services_for_cluster
from paasta_tools.utils import paasta_print


def parse_args():
    parser = argparse.ArgumentParser(
        description='Lists kubernetes instances for a service.',
    )
    parser.add_argument(
        '-c', '--cluster', dest="cluster", metavar="CLUSTER",
        default=None,
        help="define a specific cluster to read from",
    )
    parser.add_argument(
        '-d', '--soa-dir', dest="soa_dir", metavar="SOA_DIR",
        default=DEFAULT_SOA_DIR,
        help="define a different soa config directory",
    )
    args = parser.parse_args()
    return args


def main():
    args = parse_args()
    soa_dir = args.soa_dir
    cluster = args.cluster
    instances = get_services_for_cluster(
        cluster=cluster,
        instance_type='kubernetes',
        soa_dir=soa_dir,
    )
    service_instances = []
    for name, instance in instances:
        service_instances.append(compose_job_id(name, instance))
    paasta_print('\n'.join(service_instances))
    sys.exit(0)


if __name__ == "__main__":
    main()
